from __future__ import annotations

from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_pyobjects_blackbox_kernels() -> None:
    objects = [object(), None, object()]
    table = MicroPartition.from_pydict({"keys": [0, 1, 2], "objs": objects})
    # Head.
    assert table.head(2).to_pydict()["objs"] == objects[:2]
    # Filter.
    assert table.filter([col("keys") > 0]).to_pydict()["objs"] == objects[1:]


def test_nested_blackbox_kernels() -> None:
    structs = [{"a": 1, "b": 2}, None, {"a": 3}]
    lists = [[1, 2], None, [3]]
    table = MicroPartition.from_pydict({"keys": [0, 1, 2], "structs": structs, "lists": lists})
    # pyarrow fills in implicit field-internal Nones on a .to_pylist() conversion.
    structs[2]["b"] = None
    # Head.
    head_result = table.head(2).to_pydict()
    assert head_result["structs"] == structs[:2]
    assert head_result["lists"] == lists[:2]
    # Filter.
    filter_result = table.filter([col("keys") > 0]).to_pydict()
    assert filter_result["structs"] == structs[1:]
    assert filter_result["lists"] == lists[1:]



from __future__ import annotations

import argparse
import contextlib
import csv
import logging
import math
import os
import pathlib
import platform
import socket
import subprocess
import warnings
from datetime import datetime, timezone
from typing import Any, Callable, Literal

import ray

import daft
from benchmarking.tpch import answers, data_generation
from daft import DataFrame
from daft.runners.profiler import profiler

logger = logging.getLogger(__name__)

ALL_TABLES = [
    "part",
    "supplier",
    "partsupp",
    "customer",
    "orders",
    "lineitem",
    "nation",
    "region",
]


class MetricsBuilder:
    NUM_TPCH_QUESTIONS = 22

    HEADERS = [
        "started_at",
        "runner",
        "commit_hash",
        "commit_time",
        "release_tag",
        "env",
        "python_version",
        "github_runner_os",
        "github_runner_arch",
        "github_workflow",
        "github_run_id",
        "github_run_attempt",
        "github_ref",
        *[f"tpch_q{i}" for i in range(1, NUM_TPCH_QUESTIONS + 1)],
        "worker_count",
        "worker_instance_name",
    ]

    def __init__(self, runner: str):
        self._runner = runner
        self._env = "github_actions" if os.getenv("GITHUB_ACTIONS") else socket.gethostname()
        self._commit_hash = subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("utf-8").strip()
        self._commit_time = (
            subprocess.check_output(["git", "show", "-s", "--format=%cI", "HEAD"]).decode("utf-8").strip()
        )
        self._release_tag = subprocess.check_output(["git", "tag", "--points-at", "HEAD"]).decode("utf-8").strip()

        self._metrics: dict[str, Any] = {
            "started_at": datetime.now(timezone.utc).isoformat(),
            "runner": runner,
            "commit_hash": self._commit_hash,
            "commit_time": self._commit_time,
            "release_tag": self._release_tag,
            "env": self._env,
            "python_version": ".".join(platform.python_version_tuple()),
            "github_runner_os": os.getenv("RUNNER_OS"),
            "github_runner_arch": os.getenv("RUNNER_ARCH"),
            "github_workflow": os.getenv("GITHUB_WORKFLOW"),
            "github_run_id": os.getenv("GITHUB_RUN_ID"),
            "github_run_attempt": os.getenv("GITHUB_RUN_ATTEMPT"),
            "github_ref": os.getenv("GITHUB_REF"),
            "worker_count": os.getenv("WORKER_COUNT"),
            "worker_instance_name": os.getenv("WORKER_INSTANCE_NAME"),
        }

    @contextlib.contextmanager
    def collect_metrics(self, qnum: int):
        logger.info("Running benchmarks for TPC-H q %s", qnum)
        start = datetime.now()
        profile_filename = (
            f"tpch_q{qnum}_{self._runner}_{datetime.replace(start, microsecond=0).isoformat()}_viztracer.json"
        )
        with profiler(profile_filename):
            yield
        walltime_s = (datetime.now() - start).total_seconds()
        logger.info("Finished benchmarks for q%s: %ss", qnum, walltime_s)
        self._metrics[f"tpch_q{qnum}"] = walltime_s

        if str(os.getenv("RAY_PROFILING")) == str(1) and self._runner == "ray":
            profile_filename = (
                f"tpch_q{qnum}_{self._runner}_{datetime.replace(start, microsecond=0).isoformat()}_raytimeline.json"
            )
            ray.timeline(profile_filename)

    def dump_csv(self, csv_output_location: str):
        if len(self._metrics) == 0:
            logger.warning("No metrics to upload!")

        with open(csv_output_location, "w", newline="") as csvfile:
            writer = csv.writer(csvfile, delimiter=",")
            writer.writerow(MetricsBuilder.HEADERS)
            writer.writerow([self._metrics.get(header, "") for header in MetricsBuilder.HEADERS])


def get_df_with_parquet_folder(parquet_folder: str) -> Callable[[str], DataFrame]:
    def _get_df(table_name: str) -> DataFrame:
        return daft.read_parquet(os.path.join(parquet_folder, table_name, "*.parquet"))

    return _get_df


def run_all_benchmarks(
    parquet_folder: str,
    questions: list[int],
    csv_output_location: str | None,
    ray_job_dashboard_url: str | None = None,
    requirements: str | None = None,
):
    get_df = get_df_with_parquet_folder(parquet_folder)

    metrics_builder = MetricsBuilder(get_daft_benchmark_runner_name())

    for i in questions:
        # Run as a Ray Job if dashboard URL is provided
        if ray_job_dashboard_url is not None:
            from benchmarking.tpch import ray_job_runner

            working_dir = pathlib.Path(os.path.dirname(__file__))
            entrypoint = working_dir / "ray_job_runner.py"
            job_params = ray_job_runner.ray_job_params(
                parquet_folder_path=parquet_folder,
                tpch_qnum=i,
                working_dir=working_dir,
                entrypoint=entrypoint,
                runtime_env=get_ray_runtime_env(requirements),
            )

            # Run once as a warmup step
            ray_job_runner.run_on_ray(
                ray_job_dashboard_url,
                {**job_params, "submission_id": job_params["submission_id"] + "-warmup"},
            )

            # Run second time to collect metrics
            with metrics_builder.collect_metrics(i):
                ray_job_runner.run_on_ray(
                    ray_job_dashboard_url,
                    job_params,
                )

        # Run locally (potentially on a local Ray cluster)
        else:
            answer = getattr(answers, f"q{i}")
            daft_df = answer(get_df)

            with metrics_builder.collect_metrics(i):
                daft_df.collect()

    if csv_output_location:
        logger.info("Writing CSV to: %s", csv_output_location)
        metrics_builder.dump_csv(csv_output_location)
    else:
        logger.info("No CSV location specified, skipping CSV write")


def generate_parquet_data(tpch_gen_folder: str, scale_factor: float, num_parts: int) -> str:
    """Generates Parquet data and returns the path to the folder.

    Args:
        tpch_gen_folder (str): Path to the folder containing the TPCH dbgen tool and generated data
        scale_factor (float): Scale factor to run on in GB

    Returns:
        str: Path to folder containing Parquet files
    """
    csv_folder = data_generation.gen_csv_files(basedir=tpch_gen_folder, scale_factor=scale_factor, num_parts=num_parts)
    return data_generation.gen_parquet(csv_folder)


def get_daft_version() -> str:
    return daft.get_version()


def get_daft_benchmark_runner_name() -> Literal["ray"] | Literal["native"]:
    """Test utility that checks the environment variable for the runner that is being used for the benchmarking."""
    name = os.getenv("DAFT_RUNNER")
    assert name is not None, "Tests must be run with $DAFT_RUNNER env var"
    name = name.lower()

    assert name in {"ray", "native"}, f"Runner name not recognized: {name}"
    return name


def get_ray_runtime_env(requirements: str | None) -> dict:
    runtime_env = {
        "py_modules": [daft],
        "eager_install": True,
        "env_vars": {
            "DAFT_PROGRESS_BAR": "0",
            "DAFT_RUNNER": "ray",
        },
    }
    if requirements:
        runtime_env.update({"pip": requirements})
    return runtime_env


def warmup_environment(requirements: str | None, parquet_folder: str):
    """Performs necessary setup of Daft on the current benchmarking environment."""
    if get_daft_benchmark_runner_name() == "ray":
        runtime_env = get_ray_runtime_env(requirements)

        ray.init(
            runtime_env=runtime_env,
        )

        logger.info("Warming up Ray cluster with a function...")

        # NOTE: installation of runtime_env is supposed to be eager but it seems to be happening async.
        # Here we farm out some work on Ray to warm up all the workers by downloading data.
        # Warm up n := num_cpus workers.
        @ray.remote(num_cpus=1, scheduling_strategy="SPREAD")
        def warm_up_function():
            import time

            time.sleep(1)
            return get_daft_version()

        num_workers_to_warm = int(ray.cluster_resources()["CPU"])
        tasks = [warm_up_function.remote() for _ in range(num_workers_to_warm)]
        assert ray.get(tasks) == [get_daft_version() for _ in range(num_workers_to_warm)]
        del tasks

        logger.info("Ray cluster warmed up")

    get_df = get_df_with_parquet_folder(parquet_folder)
    for table in ALL_TABLES:
        df = get_df(table)
        logger.info(
            "Warming up local execution environment by loading table %s and counting rows: %s",
            table,
            df.count(df.columns[0]).to_pandas(),
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--tpch_gen_folder",
        default="data/tpch-dbgen",
        help="Path to the folder containing the TPCH dbgen tool and generated data",
    )
    parser.add_argument(
        "--parquet_file_cache",
        default=None,
        help="Path to root folder (local or in S3) containing cached Parquet files",
    )
    parser.add_argument("--scale_factor", default=10.0, help="Scale factor to run on in GB", type=float)
    parser.add_argument(
        "--num_parts", default=None, help="Number of parts to generate (defaults to 1 part per GB)", type=int
    )
    parser.add_argument("--questions", type=str, default=None, help="Comma-separated list of questions to run")
    parser.add_argument("--skip_questions", type=str, default=None, help="Comma-separated list of questions to skip")
    parser.add_argument("--output_csv", default=None, type=str, help="Location to output CSV file")
    parser.add_argument(
        "--output_csv_headers",
        action="store_true",
        help="DEPRECATED: We always output CSV headers regardless of this flag",
    )
    parser.add_argument(
        "--requirements",
        default=None,
        help="Path to pip-style requirements.txt file to bootstrap environment on remote Ray cluster",
    )
    parser.add_argument(
        "--skip_warmup",
        action="store_true",
        help="Skip warming up data before benchmark",
    )
    parser.add_argument(
        "--ray_job_dashboard_url",
        default=None,
        help="Ray Dashboard URL to submit jobs instead of using Ray client, most useful when running on a remote cluster",
    )

    args = parser.parse_args()
    if args.output_csv_headers:
        warnings.warn("Detected --output_csv_headers flag, but this flag is deprecated - CSVs always output headers")

    num_parts = math.ceil(args.scale_factor) if args.num_parts is None else args.num_parts

    # Generate Parquet data, or skip if data is cached on disk
    parquet_folder: str
    if args.parquet_file_cache is not None:
        parquet_folder = (
            os.path.join(args.parquet_file_cache, str(args.scale_factor).replace(".", "_"), str(num_parts), "parquet")
            + "/"
        )
    else:
        parquet_folder = generate_parquet_data(args.tpch_gen_folder, args.scale_factor, num_parts)

    if args.skip_warmup:
        warnings.warn("Detected --skip_warmup flag, skipping warm up task")
    else:
        warmup_environment(args.requirements, parquet_folder)

    if args.skip_questions is not None:
        if args.questions is not None:
            raise ValueError("Cannot specify both --questions and --skip_questions")
        skip_questions = {int(s) for s in args.skip_questions.split(",")}
        questions = [q for q in range(1, MetricsBuilder.NUM_TPCH_QUESTIONS + 1) if q not in skip_questions]
    elif args.questions is not None:
        questions = sorted(set(int(s) for s in args.questions.split(",")))
    else:
        questions = list(range(1, MetricsBuilder.NUM_TPCH_QUESTIONS + 1))

    run_all_benchmarks(
        parquet_folder,
        questions=questions,
        csv_output_location=args.output_csv,
        ray_job_dashboard_url=args.ray_job_dashboard_url,
        requirements=args.requirements,
    )



from __future__ import annotations

import pytest

from daft.daft import CountMode
from daft.datatype import DataType
from daft.expressions import col
from daft.recordbatch import MicroPartition


@pytest.fixture
def table():
    return MicroPartition.from_pydict({"col": [None, [], ["a"], [None], ["a", "a"], ["a", None], ["a", None, "a"]]})


@pytest.fixture
def fixed_table():
    table = MicroPartition.from_pydict({"col": [["a", "a"], ["a", "a"], ["a", None], [None, None], None]})
    fixed_dtype = DataType.fixed_size_list(DataType.string(), 2)
    return table.eval_expression_list([col("col").cast(fixed_dtype)])


def test_list_lengths(table):
    result = table.eval_expression_list([col("col").list.lengths()])
    assert result.to_pydict() == {"col": [None, 0, 1, 1, 2, 2, 3]}


def test_fixed_list_lengths(fixed_table):
    result = fixed_table.eval_expression_list([col("col").list.lengths()])
    assert result.to_pydict() == {"col": [2, 2, 2, 2, None]}


def test_list_count(table):
    result = table.eval_expression_list([col("col").list.count(CountMode.All)])
    assert result.to_pydict() == {"col": [None, 0, 1, 1, 2, 2, 3]}

    result = table.eval_expression_list([col("col").list.count(CountMode.Valid)])
    assert result.to_pydict() == {"col": [None, 0, 1, 0, 2, 1, 2]}

    result = table.eval_expression_list([col("col").list.count(CountMode.Null)])
    assert result.to_pydict() == {"col": [None, 0, 0, 1, 0, 1, 1]}


def test_fixed_list_count(fixed_table):
    result = fixed_table.eval_expression_list([col("col").list.count(CountMode.All)])
    assert result.to_pydict() == {"col": [2, 2, 2, 2, None]}

    result = fixed_table.eval_expression_list([col("col").list.count(CountMode.Valid)])
    assert result.to_pydict() == {"col": [2, 2, 1, 0, None]}

    result = fixed_table.eval_expression_list([col("col").list.count(CountMode.Null)])
    assert result.to_pydict() == {"col": [0, 0, 1, 2, None]}


def test_list_length(fixed_table):
    with pytest.warns(DeprecationWarning):
        lengths_result = fixed_table.eval_expression_list([col("col").list.lengths()])
    length_result = fixed_table.eval_expression_list([col("col").list.length()])

    assert lengths_result.to_pydict() == {"col": [2, 2, 2, 2, None]}
    assert length_result.to_pydict() == {"col": [2, 2, 2, 2, None]}



from __future__ import annotations

import pytest

from daft.datatype import DataType
from daft.expressions import col
from daft.recordbatch import MicroPartition

# Test data includes various combinations of boolean values and nulls
table = MicroPartition.from_pydict(
    {
        "a": [
            [True, True],  # all True
            [True, False],  # mixed
            [False, False],  # all False
            [True, None],  # True with null
            [False, None],  # False with null
            [None, None],  # all null
            None,  # null list
        ]
    }
)

fixed_dtype = DataType.fixed_size_list(DataType.bool(), 2)
fixed_table = table.eval_expression_list([col("a").cast(fixed_dtype)])


@pytest.mark.parametrize("table", [table, fixed_table])
def test_list_bool_and(table):
    result = table.eval_expression_list([col("a").list.bool_and()])
    assert result.to_pydict() == {"a": [True, False, False, True, False, None, None]}


@pytest.mark.parametrize("table", [table, fixed_table])
def test_list_bool_or(table):
    result = table.eval_expression_list([col("a").list.bool_or()])
    assert result.to_pydict() == {"a": [True, True, False, True, False, None, None]}



from __future__ import annotations

import pytest

import daft

from .conftest import minio_create_bucket


@pytest.mark.integration()
def test_files_roundtrip_minio_native_downloader(minio_io_config):
    bucket_name = "my-bucket"
    folder = f"s3://{bucket_name}/my-folder"
    with minio_create_bucket(minio_io_config=minio_io_config, bucket_name=bucket_name):
        bytes_data = [b"a", b"b", b"c"]
        data = {"data": bytes_data}
        df = daft.from_pydict(data)
        df = df.with_column("file_paths", df["data"].url.upload(folder, io_config=minio_io_config))
        df.collect()

        df = df.with_column("roundtrip_data", df["file_paths"].url.download(io_config=minio_io_config))
        results = df.to_pydict()

        assert results["data"] == results["roundtrip_data"] == bytes_data
        for path, expected in zip(results["file_paths"], bytes_data):
            assert path.startswith(folder)



from __future__ import annotations

import numpy as np
import pyarrow as pa

from daft import Series


def test_float_is_nan() -> None:
    s = Series.from_arrow(pa.array([1.0, np.nan, 3.0, float("nan")]))
    result = s.float.is_nan()
    assert result.to_pylist() == [False, True, False, True]


def test_float_is_nan_with_nulls() -> None:
    s = Series.from_arrow(pa.array([1.0, None, np.nan, 3.0, None, float("nan")]))
    result = s.float.is_nan()
    assert result.to_pylist() == [False, None, True, False, None, True]


def test_float_is_nan_empty() -> None:
    s = Series.from_arrow(pa.array([], type=pa.float64()))
    result = s.float.is_nan()
    assert result.to_pylist() == []


def test_float_is_nan_all_null() -> None:
    s = Series.from_arrow(pa.array([None, None, None]))
    result = s.float.is_nan()
    assert result.to_pylist() == [None, None, None]


def test_float_is_inf() -> None:
    s = Series.from_arrow(pa.array([-float("inf"), 0.0, np.inf]))
    result = s.float.is_inf()
    assert result.to_pylist() == [True, False, True]


def test_float_is_inf_with_nulls() -> None:
    s = Series.from_arrow(pa.array([-np.inf, None, 1.0, None, float("inf")]))
    result = s.float.is_inf()
    assert result.to_pylist() == [True, None, False, None, True]


def test_float_is_inf_empty() -> None:
    s = Series.from_arrow(pa.array([], type=pa.float64()))
    result = s.float.is_inf()
    assert result.to_pylist() == []


def test_float_is_inf_all_null() -> None:
    s = Series.from_arrow(pa.array([None, None, None]))
    result = s.float.is_inf()
    assert result.to_pylist() == [None, None, None]


def test_float_not_nan() -> None:
    s = Series.from_arrow(pa.array([1.0, np.nan, 3.0, float("nan")]))
    result = s.float.not_nan()
    assert result.to_pylist() == [True, False, True, False]


def test_float_not_nan_with_nulls() -> None:
    s = Series.from_arrow(pa.array([1.0, None, np.nan, 3.0, None, float("nan")]))
    result = s.float.not_nan()
    assert result.to_pylist() == [True, None, False, True, None, False]


def test_float_not_nan_empty() -> None:
    s = Series.from_arrow(pa.array([], type=pa.float64()))
    result = s.float.not_nan()
    assert result.to_pylist() == []


def test_float_not_nan_all_null() -> None:
    s = Series.from_arrow(pa.array([None, None, None]))
    result = s.float.not_nan()
    assert result.to_pylist() == [None, None, None]


def test_float_fill_nan() -> None:
    s = Series.from_arrow(pa.array([1.0, np.nan, 3.0, float("nan")]))
    fill = Series.from_arrow(pa.array([2.0]))
    result = s.float.fill_nan(fill)

    assert result.to_pylist() == [1.0, 2.0, 3.0, 2.0]


def test_float_fill_nan_with_nulls() -> None:
    s = Series.from_arrow(pa.array([1.0, None, np.nan, 3.0, None, float("nan")]))
    fill = Series.from_arrow(pa.array([2.0]))
    result = s.float.fill_nan(fill)
    assert result.to_pylist() == [1.0, None, 2.0, 3.0, None, 2.0]


def test_float_fill_nan_empty() -> None:
    s = Series.from_arrow(pa.array([], type=pa.float64()))
    fill = Series.from_arrow(pa.array([2.0]))
    result = s.float.fill_nan(fill)
    assert result.to_pylist() == []


def test_float_fill_nan_all_null() -> None:
    s = Series.from_arrow(pa.array([None, None, None]))
    fill = Series.from_arrow(pa.array([2.0]))
    result = s.float.fill_nan(fill)
    assert result.to_pylist() == [None, None, None]



from __future__ import annotations

import copy
import gc
import os
from collections import Counter
from datetime import date, datetime

import numpy as np
import psutil
import pyarrow as pa
import pytest

from daft import DataType, Series
from tests.series import ARROW_FLOAT_TYPES, ARROW_INT_TYPES, ARROW_STRING_TYPES


class CustomTestObject:
    def __init__(self, a):
        self.a = a


def test_series_arrow_array_round_trip() -> None:
    arrow = pa.array([1, 2, 3, None, 5, None])
    s = Series.from_arrow(arrow)
    back_to_arrow = s.to_arrow()
    assert arrow == back_to_arrow


def test_series_arrow_chunked_array_round_trip() -> None:
    arrow = pa.chunked_array([[1, 2, 3], [None, 5, None]])
    s = Series.from_arrow(arrow)
    back_to_arrow = s.to_arrow()
    assert arrow.combine_chunks() == back_to_arrow


@pytest.mark.parametrize("pyobj", ["allow", "disallow", "force"])
def test_series_pylist_round_trip_objects(pyobj) -> None:
    data = [1, 2, 3, 4, None]
    s = Series.from_pylist(data, pyobj=pyobj)
    back_to_list = s.to_pylist()
    assert data == back_to_list


def test_series_pyobj_wrong_parameter() -> None:
    data = [1, 2, 3, 4, None]
    with pytest.raises(ValueError):
        Series.from_pylist(data, pyobj="blah")


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES + ARROW_FLOAT_TYPES + ARROW_STRING_TYPES)
def test_series_pylist_round_trip(dtype) -> None:
    data = pa.array([1, 2, 3, None, 5, None])

    s = Series.from_arrow(data.cast(dtype))

    words = Counter(repr(s).split())
    assert s.name() in words
    assert words[s.name()] == 1
    assert words["None"] == 2


def test_series_pyobj_roundtrip() -> None:
    objects = [CustomTestObject(0), CustomTestObject(1), None]

    s = Series.from_pylist(objects)
    result = s.to_pylist()

    assert result[0] is objects[0]
    assert result[1] is objects[1]
    assert result[2] is None


def test_series_pyobj_strict_arrow_err() -> None:
    objects = [0, CustomTestObject(1)]

    with pytest.raises(pa.lib.ArrowInvalid):
        Series.from_pylist(objects, pyobj="disallow")


def test_series_pyobj_explicit_roundtrip() -> None:
    objects = [0, 1.1, "foo"]

    s = Series.from_pylist(objects, pyobj="force")

    result = s.to_pylist()

    assert result[0] is objects[0]
    assert result[1] is objects[1]
    assert result[2] is objects[2]


def test_serialize_with_pyobjects() -> None:
    objects = [CustomTestObject(1), None]

    s = Series.from_pylist(objects)
    import pickle

    s = pickle.loads(pickle.dumps(s))

    result = s.to_pylist()

    assert isinstance(result[0], CustomTestObject)
    assert result[0].a == 1
    assert result[1] is None


def test_series_pylist_round_trip_null() -> None:
    data = pa.array([None, None])

    s = Series.from_arrow(data)

    words = Counter(repr(s).split())
    assert s.name() in words
    assert words["None"] == 2


@pytest.mark.parametrize("type", [pa.binary(), pa.binary(1)])
def test_series_pylist_round_trip_binary(type) -> None:
    data = pa.array([b"a", b"b", b"c", None, b"d", None], type=type)

    s = Series.from_arrow(data)

    words = Counter(repr(s).split())
    assert s.name() in words
    assert words[s.name()] == 1
    assert words["None"] == 2


@pytest.mark.parametrize("dtype", ARROW_FLOAT_TYPES + ARROW_INT_TYPES + ARROW_STRING_TYPES)
def test_series_pickling(dtype) -> None:
    s = Series.from_pylist([1, 2, 3, None]).cast(DataType.from_arrow_type(dtype))
    copied_s = copy.deepcopy(s)
    assert s.datatype() == copied_s.datatype()
    assert s.to_pylist() == copied_s.to_pylist()


@pytest.mark.parametrize("dtype", ARROW_FLOAT_TYPES + ARROW_INT_TYPES + ARROW_STRING_TYPES)
def test_series_bincode_serdes(dtype) -> None:
    s = Series.from_pylist([1, 2, 3, None]).cast(DataType.from_arrow_type(dtype))
    serialized = s._debug_bincode_serialize()
    copied_s = Series._debug_bincode_deserialize(serialized)

    assert s.name() == copied_s.name()
    assert s.datatype() == copied_s.datatype()
    assert s.to_pylist() == copied_s.to_pylist()


def test_series_bincode_serdes_fixed_size_binary() -> None:
    s = Series.from_arrow(pa.array([b"a", b"b", b"c", None, b"d", None], type=pa.binary(1)))
    serialized = s._debug_bincode_serialize()
    copied_s = Series._debug_bincode_deserialize(serialized)

    assert s.name() == copied_s.name()
    assert s.datatype() == copied_s.datatype()
    assert s.to_pylist() == copied_s.to_pylist()


@pytest.mark.parametrize(
    "data",
    [
        [{"a": "foo", "b": "bar"}, None, {"b": "baz", "c": "quux"}],
        [[1, 2, 3], None, [4, 5]],
        [datetime.now(), None, datetime.now()],
        [date.today(), date.today(), None],
    ],
)
def test_series_bincode_serdes_on_data(data) -> None:
    s = Series.from_arrow(pa.array(data))
    serialized = s._debug_bincode_serialize()
    copied_s = Series._debug_bincode_deserialize(serialized)

    assert s.name() == copied_s.name()
    assert s.datatype() == copied_s.datatype()
    assert s.to_pylist() == copied_s.to_pylist()


def test_series_bincode_serdes_on_ext_type(uuid_ext_type) -> None:
    arr = pa.array(f"{i}".encode() for i in range(5))
    data = pa.ExtensionArray.from_storage(uuid_ext_type, arr)
    s = Series.from_arrow(data)
    serialized = s._debug_bincode_serialize()
    copied_s = Series._debug_bincode_deserialize(serialized)

    assert s.name() == copied_s.name()
    assert s.datatype() == copied_s.datatype()
    assert s.to_pylist() == copied_s.to_pylist()


def test_series_bincode_serdes_on_complex_types() -> None:
    s = Series.from_pylist([np.ones((4, 5)), np.ones((2, 10))])
    serialized = s._debug_bincode_serialize()
    copied_s = Series._debug_bincode_deserialize(serialized)

    assert s.name() == copied_s.name()
    assert s.datatype() == copied_s.datatype()
    assert all(np.all(left == right) for left, right in zip(s.to_pylist(), copied_s.to_pylist()))


def test_series_bincode_serdes_on_null_types() -> None:
    s = Series.from_pylist([None, None, None])
    serialized = s._debug_bincode_serialize()
    copied_s = Series._debug_bincode_deserialize(serialized)

    assert s.name() == copied_s.name()
    assert s.datatype() == copied_s.datatype()
    assert s.to_pylist() == copied_s.to_pylist()


def test_series_numpy_datetime64_datetime():
    from datetime import datetime

    py_datetimes = [
        datetime(2020, 10, 1, 12, 30, 45),
        datetime(2021, 10, 2, 13, 31, 46),
        datetime(2022, 10, 3, 14, 32, 47),
        datetime(2023, 10, 4, 15, 33, 48),
        datetime(2024, 10, 5, 16, 34, 49),
        datetime(2025, 10, 6, 17, 35, 50),
    ]
    for format_ in ["s", "ms", "us", "ns"]:
        list_from_datetimes = [np.datetime64(dt, format_) for dt in py_datetimes]
        # test from_pylist
        s1 = Series.from_pylist(list_from_datetimes)
        assert s1.to_pylist() == py_datetimes
        # test from_numpy
        s2 = Series.from_numpy(np.array(list_from_datetimes))
        assert s2.to_pylist() == py_datetimes


def test_series_numpy_datetime64_date():
    from datetime import datetime

    py_datetimes = [
        datetime(2020, 10, 1, 12, 30, 45),
        datetime(2021, 10, 2, 13, 31, 46),
        datetime(2022, 10, 3, 14, 32, 47),
        datetime(2023, 10, 4, 15, 33, 48),
        datetime(2024, 10, 5, 16, 34, 49),
        datetime(2025, 10, 6, 17, 35, 50),
    ]
    py_dates = [dt.date() for dt in py_datetimes]
    for format_ in ["D"]:
        # make list of numpy datetime64
        list_from_datetimes = [np.datetime64(dt, format_) for dt in py_datetimes]
        list_from_dates = [np.datetime64(d, format_) for d in py_dates]
        # test from_pylist
        s1 = Series.from_pylist(list_from_datetimes)
        s2 = Series.from_pylist(list_from_dates)
        assert s1.to_pylist() == py_dates
        assert s2.to_pylist() == py_dates
        # test from_numpy
        s3 = Series.from_numpy(np.array(list_from_datetimes))
        s4 = Series.from_numpy(np.array(list_from_dates))
        assert s3.to_pylist() == py_dates
        assert s4.to_pylist() == py_dates


def test_series_iter() -> None:
    arrow = pa.array([1, 2, 3, None, 5, None])
    s = Series.from_arrow(arrow)
    for original, iterated in zip(arrow, s):
        assert original.as_py() == iterated


def get_memory_usage() -> float:
    """Get the current memory usage in MiB."""
    process = psutil.Process(os.getpid())
    return process.memory_info().rss / (1024 * 1024)


def test_series_iter_memory_efficiency() -> None:
    # Create a Series with 1 million elements.
    n = 1000_000
    arrow = pa.array(list(range(n)))
    s = Series.from_arrow(arrow)
    sum_using_iter = 0
    sum_using_pylist = 0

    # Force garbage collection before measurements.
    gc.collect()

    # Test 1: Using iterator.
    iter_memory_before = get_memory_usage()
    iter_peak_memory = 0
    for i, item in enumerate(s):
        sum_using_iter += item or 0
        # Periodically check peak memory usage.
        if i % 100_000 == 0:
            current_memory = get_memory_usage() - iter_memory_before
            iter_peak_memory = max(iter_peak_memory, current_memory)

    current_memory = get_memory_usage() - iter_memory_before
    iter_peak_memory = max(iter_peak_memory, current_memory)

    # Force garbage collection before the next test.
    gc.collect()

    # Test 2: Using to_pylist().
    pylist_memory_before = get_memory_usage()
    pylist_peak_memory = 0
    py_list = s.to_pylist()
    for i, item in enumerate(py_list):
        sum_using_pylist += item or 0
        # Periodically check peak memory usage.
        if i % 100_000 == 0:
            current_memory = get_memory_usage() - pylist_memory_before
            pylist_peak_memory = max(pylist_peak_memory, current_memory)

    current_memory = get_memory_usage() - pylist_memory_before
    pylist_peak_memory = max(pylist_peak_memory, current_memory)

    # Check correctness.
    assert sum_using_iter == sum_using_pylist
    # Assert that the iterator uses less memory.
    assert iter_peak_memory < pylist_peak_memory, "Iterator should use less memory than to_pylist()"



from __future__ import annotations

import datetime
import io
import os
import time
from collections.abc import Iterator

import boto3
import pytest

import daft
import daft.context
from tests.conftest import get_tests_daft_runner_name
from tests.io.mock_aws_server import start_service, stop_process


@pytest.fixture(scope="session")
def aws_log_file(tmp_path_factory: pytest.TempPathFactory) -> Iterator[io.IOBase]:
    # NOTE(Clark): We have to use a log file for the mock AWS server's stdout/sterr.
    # - If we use None, then the server output will spam stdout.
    # - If we use PIPE, then the server will deadlock if the (relatively small) buffer fills, and the server is pretty
    #   noisy.
    # - If we use DEVNULL, all log output is lost.
    # With a tmp_path log file, we can prevent spam and deadlocks while also providing an avenue for debuggability, via
    # changing this fixture to something persistent, or dumping the file to stdout before closing the file, etc.
    tmp_path = tmp_path_factory.mktemp("aws_logging")
    with open(tmp_path / "aws_log.txt", "w") as f:
        yield f


def test_s3_credentials_refresh(aws_log_file: io.IOBase):
    host = "127.0.0.1"
    port = 5000

    server_url = f"http://{host}:{port}"

    bucket_name = "mybucket"
    input_file_path = f"s3://{bucket_name}/input.parquet"
    output_file_path = f"s3://{bucket_name}/output.parquet"

    old_env = os.environ.copy()
    # Set required AWS environment variables before starting server.
    # Required to opt out of concurrent writing, since we don't provide a LockClient.
    os.environ["AWS_S3_ALLOW_UNSAFE_RENAME"] = "true"

    # Start moto server.
    process = start_service(host, port, aws_log_file)

    aws_credentials = {
        "AWS_ACCESS_KEY_ID": "testing",
        "AWS_SECRET_ACCESS_KEY": "testing",
        "AWS_SESSION_TOKEN": "testing",
    }

    s3 = boto3.resource(
        "s3",
        region_name="us-west-2",
        use_ssl=False,
        endpoint_url=server_url,
        aws_access_key_id=aws_credentials["AWS_ACCESS_KEY_ID"],
        aws_secret_access_key=aws_credentials["AWS_SECRET_ACCESS_KEY"],
        aws_session_token=aws_credentials["AWS_SESSION_TOKEN"],
    )
    bucket = s3.Bucket(bucket_name)
    bucket.create(CreateBucketConfiguration={"LocationConstraint": "us-west-2"})

    count_get_credentials = 0

    def get_credentials():
        nonlocal count_get_credentials
        count_get_credentials += 1
        return daft.io.S3Credentials(
            key_id=aws_credentials["AWS_ACCESS_KEY_ID"],
            access_key=aws_credentials["AWS_SECRET_ACCESS_KEY"],
            session_token=aws_credentials["AWS_SESSION_TOKEN"],
            expiry=(datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=1)),
        )

    static_config = daft.io.IOConfig(
        s3=daft.io.S3Config(
            endpoint_url=server_url,
            region_name="us-west-2",
            key_id=aws_credentials["AWS_ACCESS_KEY_ID"],
            access_key=aws_credentials["AWS_SECRET_ACCESS_KEY"],
            session_token=aws_credentials["AWS_SESSION_TOKEN"],
            use_ssl=False,
        )
    )

    dynamic_config = daft.io.IOConfig(
        s3=daft.io.S3Config(
            endpoint_url=server_url,
            region_name="us-west-2",
            credentials_provider=get_credentials,
            buffer_time=0,
            use_ssl=False,
        )
    )

    df = daft.from_pydict({"a": [1, 2, 3]})
    df.write_parquet(input_file_path, io_config=static_config)

    df = daft.read_parquet(input_file_path, io_config=dynamic_config)
    assert count_get_credentials == 1

    df.collect()
    assert count_get_credentials == 1

    df = daft.read_parquet(input_file_path, io_config=dynamic_config)
    assert count_get_credentials == 1

    time.sleep(1)
    df.collect()
    assert count_get_credentials == 2

    df.write_parquet(output_file_path, io_config=dynamic_config)

    is_ray_runner = (
        get_tests_daft_runner_name() == "ray"
    )  # hack because ray runner will not increment `count_get_credentials`
    assert count_get_credentials == 2 or is_ray_runner

    df2 = daft.read_parquet(output_file_path, io_config=static_config)

    assert df.to_arrow() == df2.to_arrow()

    df.write_parquet(output_file_path, io_config=dynamic_config, write_mode="overwrite")
    assert count_get_credentials == 2 or is_ray_runner

    df2 = daft.read_parquet(output_file_path, io_config=static_config)

    assert df.to_arrow() == df2.to_arrow()

    time.sleep(1)
    df.write_parquet(output_file_path, io_config=dynamic_config, write_mode="overwrite")
    assert count_get_credentials == 3 or is_ray_runner

    df2 = daft.read_parquet(output_file_path, io_config=static_config)

    assert df.to_arrow() == df2.to_arrow()

    # Shutdown moto server.
    stop_process(process)
    # Restore old set of environment variables.
    os.environ = old_env



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, TypedDict

if TYPE_CHECKING:
    from daft.logical.schema import Schema
    from daft.recordbatch import MicroPartition


@dataclass(frozen=True)
class Preview:
    partition: MicroPartition | None
    total_rows: int | None


PreviewFormat = Literal[
    "fancy",
    "plain",
    "simple",
    "grid",
    "markdown",
    "html",
    # TODO "latex"
]


PreviewAlign = Literal[
    "auto",
    "left",
    "center",
    "right",
]


class PreviewColumn(TypedDict, total=False):
    header: str
    info: str
    max_width: int
    align: PreviewAlign


class PreviewOptions:
    """Preview options for show formatting.

    Usage:
        - If columns are given, their length MUST match the schema.
        - If columns are given, their settings override any global settings.

    Options:
        verbose     (bool)                      : verbose will print header info
        null        (str)                       : null string, default is 'None'
        max_width   (int)                       : global max column width
        align       (PreviewAlign)              : global column align
        columns     (list[PreviewColumn]|None)  : column overrides

    """

    _options: dict[str, object]  # normalized options

    def __init__(self, **options: Any) -> None:
        self._options = {
            "verbose": options.get("verbose", False),
            "null": options.get("null", "None"),
            "max_width": options.get("max_width", 30),
            "align": options.get("align", "left"),
            "columns": options.get("columns"),
        }

    def __repr__(self) -> str:
        """For debugging."""
        return self.serialize()

    def serialize(self) -> str:
        """This lowers the burden to interop with the rust formatter."""
        return json.dumps(self._options)


class PreviewFormatter:
    _preview: Preview
    _schema: Schema
    _format: PreviewFormat | None
    _options: PreviewOptions

    def __init__(
        self,
        preview: Preview,
        schema: Schema,
        format: PreviewFormat | None = None,
        **options: Any,
    ) -> None:
        self._preview = preview
        self._schema = schema
        self._format = format
        self._options = PreviewOptions(**options)

    def _get_user_message(self) -> str:
        if self._preview.partition is None:
            return "(No data to display: Dataframe not materialized)"
        if self._preview.total_rows == 0:
            return "(No data to display: Materialized dataframe has no rows)"
        if self._preview.total_rows is None:
            first_rows = len(self._preview.partition)
            return f"(Showing first {first_rows} rows)"
        else:
            first_rows = min(self._preview.total_rows, len(self._preview.partition))
            total_rows = self._preview.total_rows
            return f"(Showing first {first_rows} of {total_rows} rows)"

    def _repr_html_(self) -> str:
        if len(self._schema) == 0:
            return "<small>(No data to display: Dataframe has no columns)</small>"
        res = "<div>\n"
        res += self._to_html()
        res += f"\n<small>{self._get_user_message()}</small>\n</div>"
        return res

    def __repr__(self) -> str:
        if len(self._schema) == 0:
            return "(No data to display: Dataframe has no columns)"
        res = self._to_text()
        res += f"\n{self._get_user_message()}"
        return res

    def _to_html(self) -> str:
        if self._preview.partition is not None:
            return self._preview.partition.to_record_batch()._repr_html_()
        else:
            return self._schema._truncated_table_html()

    def _to_text(self) -> str:
        # give an error to hopefully avoid bug reports until we implement this
        if self._format == "html" and self._options:
            raise ValueError("Formatting options with HTML are not currently supported.")

        if self._preview.partition is not None:
            if self._format:
                return self._preview.partition.to_record_batch()._recordbatch.preview(
                    self._format, self._options.serialize()
                )
            else:
                return self._preview.partition.to_record_batch()._recordbatch.__repr__()
        else:
            return self._schema._truncated_table_string()



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import pytest

import daft

if TYPE_CHECKING:
    from daft.dataframe.dataframe import DataFrame


def _assert_sql_raise(sql: str, bindings: dict[str, DataFrame], error_msg: str) -> None:
    with pytest.raises(Exception) as excinfo:
        daft.sql(sql, **bindings).collect()

    assert error_msg in str(excinfo.value)


def _assert_df_op_raise(func: Callable, error_msg: str) -> None:
    with pytest.raises(Exception) as excinfo:
        func()

    assert error_msg in str(excinfo.value)


def test_div_floor():
    df = daft.from_pydict({"A": [1, 2, 3, 4], "B": [1.5, 2.5, 3.5, 4.5], "C": [4, 5, 6, 7]})

    actual1 = daft.sql("SELECT (A // 1) AS div_floor FROM df").collect().to_pydict()
    actual2 = df.select((daft.col("A") // 1).alias("div_floor")).collect().to_pydict()
    expected = {
        "div_floor": [1, 2, 3, 4],
    }
    assert actual1 == actual2 == expected

    actual1 = daft.sql("SELECT (A // B) AS div_floor FROM df").collect().to_pydict()
    actual2 = df.select((daft.col("A") // daft.col("B")).alias("div_floor")).collect().to_pydict()
    expected = {
        "div_floor": [0, 0, 0, 0],
    }
    assert actual1 == actual2 == expected

    actual1 = daft.sql("SELECT (B // A) AS div_floor FROM df").collect().to_pydict()
    actual2 = df.select((daft.col("B") // daft.col("A")).alias("div_floor")).collect().to_pydict()
    expected = {
        "div_floor": [1.0, 1.0, 1.0, 1.0],
    }
    assert actual1 == actual2 == expected

    actual1 = daft.sql("SELECT (C // A) AS div_floor FROM df").collect().to_pydict()
    actual2 = df.select((daft.col("C") // daft.col("A")).alias("div_floor")).collect().to_pydict()
    expected = {
        "div_floor": [4, 2, 2, 1],
    }
    assert actual1 == actual2 == expected


def test_unsupported_div_floor():
    df = daft.from_pydict({"A": [1, 2, 3, 4], "B": [1.5, 2.5, 3.5, 4.5], "C": [True, False, True, True]})

    bindings = {"df": df}

    _assert_sql_raise(
        "SELECT A // C FROM df", bindings, "TypeError Cannot perform floor divide on types: Int64, Boolean"
    )

    _assert_sql_raise(
        "SELECT C // A FROM df", bindings, "TypeError Cannot perform floor divide on types: Boolean, Int64"
    )

    _assert_sql_raise(
        "SELECT B // C FROM df", bindings, "TypeError Cannot perform floor divide on types: Float64, Boolean"
    )

    _assert_sql_raise(
        "SELECT B // C FROM df", bindings, "TypeError Cannot perform floor divide on types: Float64, Boolean"
    )

    _assert_df_op_raise(
        lambda: df.select(daft.col("A") // daft.col("C")).collect(),
        "Cannot perform floor divide on types: Int64, Boolean",
    )

    _assert_df_op_raise(
        lambda: df.select(daft.col("C") // daft.col("A")).collect(),
        "Cannot perform floor divide on types: Boolean, Int64",
    )

    _assert_df_op_raise(
        lambda: df.select(daft.col("B") // daft.col("C")).collect(),
        "Cannot perform floor divide on types: Float64, Boolean",
    )

    _assert_df_op_raise(
        lambda: df.select(daft.col("C") // daft.col("B")).collect(),
        "Cannot perform floor divide on types: Boolean, Float64",
    )



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from collections.abc import Iterator


from daft import DataType
from daft.api_annotations import PublicAPI
from daft.io.source import DataSource, DataSourceTask
from daft.recordbatch import MicroPartition
from daft.schema import Schema

if TYPE_CHECKING:
    from daft.dataframe import DataFrame
    from daft.io.pushdowns import Pushdowns


@overload
def _range(end: int) -> DataFrame: ...


@overload
def _range(start: int, end: int) -> DataFrame: ...


@overload
def _range(start: int, end: int, step: int) -> DataFrame: ...


@overload
def _range(start: int, end: int, step: int, partitions: int) -> DataFrame: ...


# TODO: consider using `from_range` and `Series.from_range` instead.
@PublicAPI  # type: ignore
def _range(start: int, end: int | None = None, step: int = 1, partitions: int = 1) -> DataFrame:
    """Creates a DataFrame with a range of values.

    Args:
        start (int): The start of the range.
        end (int, optional): The end of the range. If not provided, the start is 0 and the end is `start`.
        step (int, optional): The step size of the range. Defaults to 1.
        partitions (int, optional): The number of partitions to split the range into. Defaults to 1.

    Examples:
        The range starts at 0 and ends at `end` (exclusive) with a step size of 1.

        >>> import daft
        >>> daft.range(5).show()
        ╭───────╮
        │ id    │
        │ ---   │
        │ Int64 │
        ╞═══════╡
        │ 0     │
        ├╌╌╌╌╌╌╌┤
        │ 1     │
        ├╌╌╌╌╌╌╌┤
        │ 2     │
        ├╌╌╌╌╌╌╌┤
        │ 3     │
        ├╌╌╌╌╌╌╌┤
        │ 4     │
        ╰───────╯
        <BLANKLINE>
        (Showing first 5 of 5 rows)

        The range starts at `start` and ends at `end` (exclusive) with a step size of 1.

        >>> import daft
        >>> daft.range(2, 5).show()
        ╭───────╮
        │ id    │
        │ ---   │
        │ Int64 │
        ╞═══════╡
        │ 2     │
        ├╌╌╌╌╌╌╌┤
        │ 3     │
        ├╌╌╌╌╌╌╌┤
        │ 4     │
        ╰───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

        The range starts at `start` and ends at `end` (exclusive) with a step size of `step`.

        >>> import daft
        >>> daft.range(2, 10, 2).show()
        ╭───────╮
        │ id    │
        │ ---   │
        │ Int64 │
        ╞═══════╡
        │ 2     │
        ├╌╌╌╌╌╌╌┤
        │ 4     │
        ├╌╌╌╌╌╌╌┤
        │ 6     │
        ├╌╌╌╌╌╌╌┤
        │ 8     │
        ╰───────╯
        <BLANKLINE>
        (Showing first 4 of 4 rows)

        The range starts at `start` and ends at `end` (exclusive) with a step size of `step`.
        The range is partitioned into `partitions` partitions.

        >>> import daft
        >>> df = daft.range(2, 10, step=2, partitions=2)
        >>> df.num_partitions()
        2
        >>> df.show()
        ╭───────╮
        │ id    │
        │ ---   │
        │ Int64 │
        ╞═══════╡
        │ 2     │
        ├╌╌╌╌╌╌╌┤
        │ 4     │
        ├╌╌╌╌╌╌╌┤
        │ 6     │
        ├╌╌╌╌╌╌╌┤
        │ 8     │
        ╰───────╯
        <BLANKLINE>
        (Showing first 4 of 4 rows)
    """
    if end is None:
        end = start
        start = 0
    else:
        start = start
    return RangeSource(start, end, step, partitions).read()


class RangeSource(DataSource):
    """RangeSource produces a DataFrame from a range with a given step size."""

    _start: int
    _end: int
    _step: int
    _partitions: int
    _schema = Schema.from_pydict({"id": DataType.int64()})

    def __init__(self, start: int, end: int, step: int = 1, partitions: int = 1) -> None:
        """Create a RangeSource instance.

        Args:
            start (int): The start of the range.
            end (int, optional): The end of the range. If not provided, the start is 0 and the end is `start`.
            step (int, optional): The step size of the range. Defaults to 1.
            partitions (int, optional): The number of partitions to split the range into. Defaults to 1.
        """
        self._start = start
        self._end = end
        self._step = step
        self._partitions = partitions

    @property
    def name(self) -> str:
        return "RangeSource"

    @property
    def schema(self) -> Schema:
        return self._schema

    def get_tasks(self, pushdowns: Pushdowns) -> Iterator[RangeSourceTask]:
        step = self._step
        size = (self._end - self._start) // self._partitions
        curr_s = self._start
        curr_e = self._start + size
        while curr_e <= self._end:
            yield RangeSourceTask(curr_s, curr_e, step)
            # update to the next chunk
            curr_s = curr_e + step
            curr_e = curr_s + size


@dataclass
class RangeSourceTask(DataSourceTask):
    _start: int
    _end: int
    _step: int

    @property
    def schema(self) -> Schema:
        return RangeSource._schema

    def get_micro_partitions(self) -> Iterator[MicroPartition]:
        yield MicroPartition.from_pydict({"id": list(range(self._start, self._end, self._step))})



from __future__ import annotations

import uuid

import pytest
import s3fs

import daft


@pytest.fixture(scope="function")
def bucket(minio_io_config):
    # For some reason s3fs is having trouble cleaning up MinIO
    # folders created by pyarrow write_parquet. We just write to
    # paths with random UUIDs to work around this.
    BUCKET = "my-bucket"

    fs = s3fs.S3FileSystem(
        key=minio_io_config.s3.key_id,
        password=minio_io_config.s3.access_key,
        client_kwargs={"endpoint_url": minio_io_config.s3.endpoint_url},
    )
    if not fs.exists(BUCKET):
        fs.mkdir(BUCKET)
    yield BUCKET


@pytest.mark.integration()
@pytest.mark.parametrize("protocol", ["s3://", "s3a://", "s3n://"])
def test_writing_parquet(minio_io_config, bucket, protocol):
    data = {
        "foo": [1, 2, 3],
        "bar": ["a", "b", "c"],
    }
    df = daft.from_pydict(data)
    df = df.repartition(2)
    results = df.write_parquet(
        f"{protocol}{bucket}/parquet-writes-{uuid.uuid4()}",
        partition_cols=["bar"],
        io_config=minio_io_config,
    )
    results.collect()
    assert len(results) == 3



from __future__ import annotations

import dataclasses
import logging
import os
import pathlib
import sys
import urllib.parse
from datetime import datetime, timezone
from typing import Any

from daft.convert import from_pydict
from daft.daft import FileFormat, FileInfos, IOConfig, io_glob
from daft.dependencies import fsspec, pafs
from daft.expressions.expressions import col
from daft.recordbatch import MicroPartition

logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class PyArrowFSWithExpiry:
    fs: pafs.FileSystem
    expiry: datetime | None


_CACHED_FSES: dict[tuple[str, IOConfig | None], PyArrowFSWithExpiry] = {}


def _get_fs_from_cache(protocol: str, io_config: IOConfig | None) -> pafs.FileSystem | None:
    """Get an instantiated pyarrow filesystem from the cache based on the URI protocol.

    Returns None if no such cache entry exists.
    """
    global _CACHED_FSES

    if (protocol, io_config) in _CACHED_FSES:
        fs = _CACHED_FSES[(protocol, io_config)]

        if fs.expiry is None or fs.expiry > datetime.now(timezone.utc):
            return fs.fs

    return None


def _put_fs_in_cache(protocol: str, fs: pafs.FileSystem, io_config: IOConfig | None, expiry: datetime | None) -> None:
    """Put pyarrow filesystem in cache under provided protocol."""
    global _CACHED_FSES

    _CACHED_FSES[(protocol, io_config)] = PyArrowFSWithExpiry(fs, expiry)


def get_filesystem(protocol: str, **kwargs: Any) -> fsspec.AbstractFileSystem:
    if protocol == "s3" or protocol == "s3a":
        try:
            import botocore.session
        except ImportError:
            logger.error(
                "Error when importing botocore. install daft[aws] for the required 3rd party dependencies to interact with AWS S3 (https://www.getdaft.io/projects/docs/en/latest/learn/install.html)"
            )
            raise

        s3fs_kwargs = {}

        # If accessing S3 without credentials, use anonymous access: https://github.com/Eventual-Inc/Daft/issues/503
        credentials_available = botocore.session.get_session().get_credentials() is not None
        if not credentials_available:
            logger.warning(
                "AWS credentials not found - using anonymous access to S3 which will fail if the bucket you are accessing is not a public bucket. See boto3 documentation for more details on configuring your AWS credentials: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html"
            )
            s3fs_kwargs["anon"] = True

        kwargs = {**kwargs, **s3fs_kwargs}

    try:
        klass = fsspec.get_filesystem_class(protocol)
    except ImportError:
        logger.error(
            "Error when importing dependencies for accessing data with: %s. Please ensure that daft was installed with the appropriate extra dependencies (https://www.getdaft.io/projects/docs/en/latest/learn/install.html)",
            protocol,
        )
        raise

    fs = klass(**kwargs)
    return fs


def get_protocol_from_path(path: str) -> str:
    parsed_scheme = urllib.parse.urlparse(path, allow_fragments=False).scheme
    parsed_scheme = parsed_scheme.lower()
    if parsed_scheme == "" or parsed_scheme is None:
        return "file"
    if sys.platform == "win32" and len(parsed_scheme) == 1 and ("a" <= parsed_scheme) and (parsed_scheme <= "z"):
        return "file"
    return parsed_scheme


_CANONICAL_PROTOCOLS = {
    "gcs": "gs",
    "https": "http",
    "s3a": "s3",
    "ssh": "sftp",
    "arrow_hdfs": "hdfs",
    "az": "abfs",
    "blockcache": "cached",
    "jlab": "jupyter",
}


def canonicalize_protocol(protocol: str) -> str:
    """Return the canonical protocol from the provided protocol, such that there's a 1:1 mapping between protocols and pyarrow/fsspec filesystem implementations."""
    return _CANONICAL_PROTOCOLS.get(protocol, protocol)


def _resolve_paths_and_filesystem(
    paths: str | pathlib.Path | list[str],
    io_config: IOConfig | None = None,
) -> tuple[list[str], pafs.FileSystem]:
    """Resolves and normalizes the provided path and infers it's filesystem.

    Also ensures that the inferred filesystem is compatible with the passed filesystem, if provided.

    Args:
        paths: A single file/directory path or a list of file/directory paths.
            A list of paths can contain both files and directories.
        io_config: A Daft IOConfig that should be best-effort applied onto the returned
            FileSystem
    """
    if isinstance(paths, pathlib.Path):
        paths = str(paths)
    if isinstance(paths, str):
        paths = [paths]
    assert isinstance(paths, list), paths
    assert all(isinstance(p, str) for p in paths), paths
    assert len(paths) > 0, paths

    # Sanitize s3a/s3n protocols, which are produced by Hadoop-based systems as a way of denoting which s3
    # filesystem client to use. However this doesn't matter for Daft, and PyArrow cannot recognize these protocols.
    paths = [f"s3://{p[6:]}" if p.startswith("s3a://") or p.startswith("s3n://") else p for p in paths]

    # Ensure that protocols for all paths are consistent, i.e. that they would map to the
    # same filesystem.
    protocols = {get_protocol_from_path(path) for path in paths}
    canonicalized_protocols = {canonicalize_protocol(protocol) for protocol in protocols}
    if len(canonicalized_protocols) > 1:
        raise ValueError(
            "All paths must have the same canonical protocol to ensure that they are all "
            f"hitting the same storage backend, but got protocols {protocols} with canonical "
            f"protocols - {canonicalized_protocols} and full paths - {paths}"
        )

    # Canonical protocol shared by all paths.
    protocol = next(iter(canonicalized_protocols))

    # Try to get filesystem from protocol -> fs cache.
    resolved_filesystem = _get_fs_from_cache(protocol, io_config)
    if resolved_filesystem is None:
        # Resolve path and filesystem for the first path.
        # We use this first resolved filesystem for validation on all other paths.
        resolved_path, resolved_filesystem, expiry = _infer_filesystem(paths[0], io_config)

        # Put resolved filesystem in cache under these paths' canonical protocol.
        _put_fs_in_cache(protocol, resolved_filesystem, io_config, expiry)
    else:
        resolved_path = _validate_filesystem(paths[0], resolved_filesystem, io_config)

    # filesystem should be a non-None pyarrow FileSystem at this point, either
    # user-provided, taken from the cache, or inferred from the first path.
    assert resolved_filesystem is not None and isinstance(resolved_filesystem, pafs.FileSystem)

    # Resolve all other paths and validate with the user-provided/cached/inferred filesystem.
    resolved_paths = [resolved_path]
    for path in paths[1:]:
        resolved_path = _validate_filesystem(path, resolved_filesystem, io_config)
        resolved_paths.append(resolved_path)

    return resolved_paths, resolved_filesystem


def _validate_filesystem(path: str, fs: pafs.FileSystem, io_config: IOConfig | None) -> str:
    resolved_path, inferred_fs, _ = _infer_filesystem(path, io_config)
    if not isinstance(fs, type(inferred_fs)):
        raise RuntimeError(
            f"Cannot read multiple paths with different inferred PyArrow filesystems. Expected: {fs} but received: {inferred_fs}"
        )
    return resolved_path


def _infer_filesystem(
    path: str,
    io_config: IOConfig | None,
) -> tuple[str, pafs.FileSystem, datetime | None]:
    """Resolves and normalizes the provided path and infers its filesystem and expiry.

    Also ensures that the inferred filesystem is compatible with the passedfilesystem, if provided.

    Args:
        path: A single file/directory path.
        io_config: A Daft IOConfig that should be best-effort applied onto the returned
            FileSystem
    """
    protocol = get_protocol_from_path(path)
    translated_kwargs: dict[str, Any]

    def _set_if_not_none(kwargs: dict[str, Any], key: str, val: Any | None) -> None:
        """Helper method used when setting kwargs for pyarrow."""
        if val is not None:
            kwargs[key] = val

    ###
    # S3
    ###
    if protocol == "s3":
        translated_kwargs = {}
        if io_config is not None and io_config.s3 is not None:
            s3_config = io_config.s3
            _set_if_not_none(translated_kwargs, "endpoint_override", s3_config.endpoint_url)
            _set_if_not_none(translated_kwargs, "access_key", s3_config.key_id)
            _set_if_not_none(translated_kwargs, "secret_key", s3_config.access_key)
            _set_if_not_none(translated_kwargs, "session_token", s3_config.session_token)
            _set_if_not_none(translated_kwargs, "region", s3_config.region_name)
            _set_if_not_none(translated_kwargs, "anonymous", s3_config.anonymous)
            if s3_config.num_tries is not None:
                try:
                    from pyarrow.fs import AwsStandardS3RetryStrategy

                    translated_kwargs["retry_strategy"] = AwsStandardS3RetryStrategy(max_attempts=s3_config.num_tries)
                except ImportError:
                    pass  # Config does not exist in pyarrow 7.0.0

            expiry = None
            if (s3_creds := s3_config.provide_cached_credentials()) is not None:
                _set_if_not_none(translated_kwargs, "access_key", s3_creds.key_id)
                _set_if_not_none(translated_kwargs, "secret_key", s3_creds.access_key)
                _set_if_not_none(translated_kwargs, "session_token", s3_creds.session_token)

                expiry = s3_creds.expiry

        resolved_filesystem = pafs.S3FileSystem(**translated_kwargs)
        resolved_path = resolved_filesystem.normalize_path(_unwrap_protocol(path))
        return resolved_path, resolved_filesystem, expiry

    ###
    # Local
    ###
    elif protocol == "file":
        resolved_filesystem = pafs.LocalFileSystem()
        path = os.path.expanduser(path)
        resolved_path = resolved_filesystem.normalize_path(_unwrap_protocol(path))
        return resolved_path, resolved_filesystem, None

    ###
    # GCS
    ###
    elif protocol in {"gs", "gcs"}:
        try:
            from pyarrow.fs import GcsFileSystem
        except ImportError:
            raise ImportError(
                "Unable to import GcsFileSystem from pyarrow - please ensure you have pyarrow >= 9.0 or consider "
                "using Daft's native GCS IO code instead"
            )

        translated_kwargs = {}
        if io_config is not None and io_config.gcs is not None:
            gcs_config = io_config.gcs
            _set_if_not_none(translated_kwargs, "anonymous", gcs_config.anonymous)
            _set_if_not_none(translated_kwargs, "project_id", gcs_config.project_id)

        resolved_filesystem = GcsFileSystem(**translated_kwargs)
        resolved_path = resolved_filesystem.normalize_path(_unwrap_protocol(path))
        return resolved_path, resolved_filesystem, None

    ###
    # HTTP: Use FSSpec as a fallback
    ###
    elif protocol in {"http", "https"}:
        fsspec_fs_cls = fsspec.get_filesystem_class(protocol)
        fsspec_fs = fsspec_fs_cls()
        resolved_filesystem, resolved_path = pafs._resolve_filesystem_and_path(path, fsspec_fs)
        resolved_path = resolved_filesystem.normalize_path(resolved_path)
        return resolved_path, resolved_filesystem, None

    ###
    # Azure: Use FSSpec as a fallback
    ###
    elif protocol in {"az", "abfs", "abfss"}:
        fsspec_fs_cls = fsspec.get_filesystem_class(protocol)

        if io_config is not None:
            # TODO: look into support for other AzureConfig parameters
            fsspec_fs = fsspec_fs_cls(
                account_name=io_config.azure.storage_account,
                account_key=io_config.azure.access_key,
                sas_token=io_config.azure.sas_token,
                tenant_id=io_config.azure.tenant_id,
                client_id=io_config.azure.client_id,
                client_secret=io_config.azure.client_secret,
                anon=io_config.azure.anonymous,
            )
        else:
            fsspec_fs = fsspec_fs_cls()
        resolved_filesystem, resolved_path = pafs._resolve_filesystem_and_path(path, fsspec_fs)
        resolved_path = resolved_filesystem.normalize_path(_unwrap_protocol(resolved_path))
        return resolved_path, resolved_filesystem, None

    else:
        raise NotImplementedError(f"Cannot infer PyArrow filesystem for protocol {protocol}: please file an issue!")


def _unwrap_protocol(path: str) -> str:
    """Slice off any protocol prefixes on path."""
    parsed = urllib.parse.urlparse(path, allow_fragments=False)  # support '#' in path
    query = "?" + parsed.query if parsed.query else ""  # support '?' in path
    return parsed.netloc + parsed.path + query


###
# File globbing
###


def glob_path_with_stats(
    path: str,
    file_format: FileFormat | None,
    io_config: IOConfig | None,
) -> FileInfos:
    """Glob a path, returning a FileInfos."""
    files = io_glob(path, io_config=io_config)
    filepaths_to_infos = {f["path"]: {"size": f["size"], "type": f["type"]} for f in files}

    # Set number of rows if available.
    if file_format is not None and file_format == FileFormat.Parquet:
        parquet_statistics = MicroPartition.read_parquet_statistics(
            list(filepaths_to_infos.keys()),
            io_config=io_config,
        ).to_pydict()
        for path, num_rows in zip(parquet_statistics["uris"], parquet_statistics["row_count"]):
            filepaths_to_infos[path]["rows"] = num_rows

    file_paths = []
    file_sizes = []
    num_rows = []
    for path, infos in filepaths_to_infos.items():
        file_paths.append(path)
        file_sizes.append(infos.get("size"))
        num_rows.append(infos.get("rows"))

    return FileInfos.from_infos(file_paths=file_paths, file_sizes=file_sizes, num_rows=num_rows)


###
# Path joining
###


def join_path(fs: pafs.FileSystem, base_path: str, *sub_paths: str) -> str:
    """Join a base path with sub-paths using the appropriate path separator for the given filesystem."""
    if isinstance(fs, pafs.LocalFileSystem):
        return os.path.join(base_path, *sub_paths)
    else:
        return f"{base_path.rstrip('/')}/{'/'.join(sub_paths)}"


def overwrite_files(
    written_file_paths: list[str],
    root_dir: str | pathlib.Path,
    io_config: IOConfig | None,
    overwrite_partitions: bool,
) -> None:
    [resolved_path], fs = _resolve_paths_and_filesystem(root_dir, io_config=io_config)

    all_file_paths = []
    if overwrite_partitions:
        # Get all files in ONLY the directories that were written to.

        written_dirs = set(str(pathlib.Path(path).parent) for path in written_file_paths)
        for dir in written_dirs:
            file_selector = pafs.FileSelector(dir, recursive=True)
            try:
                all_file_paths.extend(
                    [info.path for info in fs.get_file_info(file_selector) if info.type == pafs.FileType.File]
                )
            except FileNotFoundError:
                continue
    else:
        # Get all files in the root directory.

        file_selector = pafs.FileSelector(resolved_path, recursive=True)
        try:
            all_file_paths.extend(
                [info.path for info in fs.get_file_info(file_selector) if info.type == pafs.FileType.File]
            )
        except FileNotFoundError:
            # The root directory does not exist, so there are no files to delete.
            return

    all_file_paths_df = from_pydict({"path": all_file_paths})

    # Find the files that were not written to in this run and delete them.
    to_delete = all_file_paths_df.where(~(col("path").is_in(written_file_paths)))

    # TODO: Look into parallelizing this
    for entry in to_delete:
        fs.delete_file(entry["path"])



from __future__ import annotations

from typing import Any

import pytest

import daft
from daft import DataFrame, col
from tests.conftest import check_answer


def helper(make_df, op: str, left: dict[str, Any], right: dict[str, Any], expected: dict[str, Any]):
    df1 = make_df(left)
    df2 = make_df(right)
    df_helper(op, df1, df2, expected)


def df_helper(op: str, df1: DataFrame, df2: DataFrame, expected: dict[str, Any]):
    if op == "intersect":
        result = df1.intersect(df2)
    elif op == "except_distinct":
        result = df1.except_distinct(df2)
    elif op == "intersect_all":
        result = df1.intersect_all(df2)
    else:
        result = df1.except_all(df2)
    check_answer(result, expected)


@pytest.mark.parametrize(
    "op, left, right, expected",
    [
        ("intersect", {"foo": [1, 2, 3]}, {"bar": [2, 3, 4]}, {"foo": [2, 3]}),
        ("intersect_all", {"foo": [1, 2, 2]}, {"bar": [2, 2, 4]}, {"foo": [2, 2]}),
        ("except_distinct", {"foo": [1, 2, 3]}, {"bar": [2, 3, 4]}, {"foo": [1]}),
        ("except_all", {"foo": [1, 2, 2]}, {"bar": [2, 4]}, {"foo": [1, 2]}),
    ],
)
def test_simple_intersect_or_except(make_df, op, left, right, expected):
    helper(make_df, op, left, right, expected)


@pytest.mark.parametrize(
    "op, left, right, expected",
    [
        ("intersect", {"foo": [1, 2, 2, 3]}, {"bar": [2, 3, 3]}, {"foo": [2, 3]}),
        ("intersect_all", {"foo": [1, 2, 2, 3]}, {"bar": [2, 3, 3]}, {"foo": [2, 3]}),
        ("except_distinct", {"foo": [1, 2, 2, 3]}, {"bar": [2, 3, 3]}, {"foo": [1]}),
        ("except_all", {"foo": [1, 2, 2, 3]}, {"bar": [2, 3, 3]}, {"foo": [1, 2]}),
    ],
)
def test_with_duplicate(make_df, op, left, right, expected):
    helper(make_df, op, left, right, expected)


@pytest.mark.parametrize(
    "op, df, expected",
    [
        ("intersect", {"foo": [1, 2, 3]}, {"foo": [1, 2, 3]}),
        ("intersect_all", {"foo": [1, 2, 3]}, {"foo": [1, 2, 3]}),
        ("except_distinct", {"foo": [1, 2, 3]}, {"foo": []}),
        ("except_all", {"foo": [1, 2, 2]}, {"foo": []}),
    ],
)
def test_with_self(make_df, op, df, expected):
    df = make_df(df)
    df_helper(op, df, df, expected)


@pytest.mark.parametrize(
    "op, left, expected",
    [
        ("intersect", {"foo": [1, 2, 3]}, {"foo": []}),
        ("intersect_all", {"foo": [1, 2, 3]}, {"foo": []}),
        ("except_distinct", {"foo": [1, 2, 3]}, {"foo": [1, 2, 3]}),
        ("except_all", {"foo": [1, 2, 2]}, {"foo": [1, 2, 2]}),
    ],
)
def test_with_empty(make_df, op, left, expected):
    df1 = make_df(left)
    df2 = make_df({"bar": []}).select(col("bar").cast(daft.DataType.int64()))
    df_helper(op, df1, df2, expected)


@pytest.mark.parametrize(
    "op, left, right, expected",
    [
        ("intersect", {"foo": [1, 2, None]}, {"foo": [2, 3, None]}, {"foo": [2, None]}),
        ("intersect_all", {"foo": [1, 2, None]}, {"foo": [2, 3, None]}, {"foo": [2, None]}),
        ("intersect", {"foo": [1, 2]}, {"foo": [2, 3, None]}, {"foo": [2]}),
        ("intersect_all", {"foo": [1, 2]}, {"foo": [2, 3, None]}, {"foo": [2]}),
        ("intersect", {"foo": [1, 2, None]}, {"foo": [2, 3]}, {"foo": [2]}),
        ("intersect_all", {"foo": [1, 2, None]}, {"foo": [2, 3]}, {"foo": [2]}),
    ],
)
def test_intersect_with_nulls(make_df, op, left, right, expected):
    helper(make_df, op, left, right, expected)


@pytest.mark.parametrize(
    "op, left, right, expected",
    [
        ("except_distinct", {"foo": [1, 2, None]}, {"foo": [2, 3, None]}, {"foo": [1]}),
        ("except_all", {"foo": [1, 2, None]}, {"foo": [2, 3, None]}, {"foo": [1]}),
        ("except_distinct", {"foo": [1, 2]}, {"foo": [2, 3, None]}, {"foo": [1]}),
        ("except_all", {"foo": [1, 2]}, {"foo": [2, 3, None]}, {"foo": [1]}),
        ("except_distinct", {"foo": [1, 2, None]}, {"foo": [2, 3]}, {"foo": [1, None]}),
        ("except_all", {"foo": [1, 2, None]}, {"foo": [2, 3]}, {"foo": [1, None]}),
    ],
)
def test_except_with_nulls(make_df, op, left, right, expected):
    helper(make_df, op, left, right, expected)


@pytest.mark.parametrize(
    "op, left, right, expected",
    [
        (
            "intersect_all",
            {"foo": [1, 2, 2], "bar": [2, 3, 3]},
            {"a": [2, 2, 4], "b": [3, 3, 4]},
            {"foo": [2, 2], "bar": [3, 3]},
        ),
        (
            "except_all",
            {"foo": [1, 2, 2], "bar": [2, 3, 3]},
            {"a": [2, 2, 4], "b": [3, 3, 4]},
            {"foo": [1], "bar": [2]},
        ),
    ],
)
def test_multiple_fields(make_df, op, left, right, expected):
    helper(make_df, op, left, right, expected)


def test_union():
    df1 = daft.from_pydict({"x": [1, 2, 3], "y": [4, 5, 6]})
    df2 = daft.from_pydict({"x": [3, 4, 5], "y": [6, 7, 8]})
    expected = {"x": [1, 2, 3, 4, 5], "y": [4, 5, 6, 7, 8]}
    actual = df1.union(df2).sort("x").to_pydict()
    actual_sql = daft.sql("select * from df1 union select * from df2").sort("x").to_pydict()
    assert actual == expected
    assert actual_sql == expected


@pytest.mark.parametrize("quantifier", [None, "all"])
def test_union_fails_for_different_schemas(quantifier):
    df1 = daft.from_pydict({"x": [1, 2, 3], "y": [4, 5, 6]})
    df2 = daft.from_pydict({"x": [3, 4, 5], "z": [6, 7, 8]})
    with pytest.raises(daft.exceptions.DaftCoreException):
        if quantifier == "all":
            df1.union_all(df2).collect()
        else:
            df1.union(df2).collect()


def test_union_all():
    df1 = daft.from_pydict({"x": [1, 2, 3], "y": [4, 5, 6]})
    df2 = daft.from_pydict({"x": [3, 2, 1], "y": [6, 5, 4]})
    expected = {"x": [1, 1, 2, 2, 3, 3], "y": [4, 4, 5, 5, 6, 6]}
    actual = df1.union_all(df2).sort("x").to_pydict()
    actual_sql = daft.sql("select * from df1 union all select * from df2").sort("x").to_pydict()
    assert actual == expected
    assert actual_sql == expected


def test_union_by_name():
    df1 = daft.from_pydict({"name": ["Alice", "Bob", "Bob"], "age": [25, 30, 30], "city": ["NY", "LA", "LA"]})

    df2 = daft.from_pydict(
        {
            "name": ["Bob", "Carol", "Bob"],
            "salary": [50000, 60000, 50000],
            "city": ["LA", "SF", "LA"],
            "occupation": ["barista", "cashier", "barista"],
        }
    )

    expected = {
        "name": ["Alice", "Bob", "Bob", "Carol"],
        "age": [25, None, 30, None],
        "city": ["NY", "LA", "LA", "SF"],
        "salary": [None, 50000, None, 60000],
        "occupation": [None, "barista", None, "cashier"],
    }

    actual_sql = (
        daft.sql("select * from df1 union by name select * from df2")
        .sort(by=["name", "city", "occupation"])
        .to_pydict()
    )
    actual_df = df1.union_by_name(df2).sort(by=["name", "city", "occupation"]).to_pydict()

    assert actual_sql == expected
    assert actual_df == expected


def test_union_all_by_name():
    df1 = daft.from_pydict({"name": ["Alice", "Bob", "Bob"], "age": [25, 30, 30], "city": ["NY", "LA", "LA"]})

    df2 = daft.from_pydict(
        {
            "name": ["Bob", "Carol", "Bob"],
            "salary": [50000, 60000, 50000],
            "city": ["LA", "SF", "LA"],
            "occupation": ["barista", "cashier", "barista"],
        }
    )

    expected = {
        "name": ["Alice", "Bob", "Bob", "Bob", "Bob", "Carol"],
        "age": [25, None, None, 30, 30, None],
        "city": ["NY", "LA", "LA", "LA", "LA", "SF"],
        "salary": [None, 50000, 50000, None, None, 60000],
        "occupation": [None, "barista", "barista", None, None, "cashier"],
    }

    actual_sql = (
        daft.sql("select * from df1 union all by name select * from df2")
        .sort(by=["name", "city", "occupation"])
        .to_pydict()
    )
    actual_df = df1.union_all_by_name(df2).sort(by=["name", "city", "occupation"]).to_pydict()

    assert actual_sql == expected
    assert actual_df == expected



from __future__ import annotations

import pytest

pyiceberg = pytest.importorskip("pyiceberg")

import contextlib

import pyarrow as pa

import daft
from tests.conftest import assert_df_equals


@contextlib.contextmanager
def table_written_by_pyiceberg(local_pyiceberg_catalog):
    schema = pa.schema([("col", pa.int64()), ("mapCol", pa.map_(pa.int32(), pa.string()))])

    data = {"col": [1, 2, 3], "mapCol": [[(1, "foo"), (2, "bar")], [(3, "baz")], [(4, "foobar")]]}
    arrow_table = pa.Table.from_pydict(data, schema=schema)
    table_name = "pyiceberg.map_table"
    try:
        table = local_pyiceberg_catalog.create_table(table_name, schema=schema)
        table.append(arrow_table)
        yield table_name
    except Exception as e:
        raise e
    finally:
        local_pyiceberg_catalog.drop_table(table_name)


@contextlib.contextmanager
def table_written_by_daft(local_pyiceberg_catalog):
    schema = pa.schema([("col", pa.int64()), ("mapCol", pa.map_(pa.int32(), pa.string()))])

    data = {"col": [1, 2, 3], "mapCol": [[(1, "foo"), (2, "bar")], [(3, "baz")], [(4, "foobar")]]}
    arrow_table = pa.Table.from_pydict(data, schema=schema)
    table_name = "pyiceberg.map_table"
    try:
        table = local_pyiceberg_catalog.create_table(table_name, schema=schema)
        df = daft.from_arrow(arrow_table)
        df.write_iceberg(table, mode="overwrite")
        table.refresh()
        yield table_name
    except Exception as e:
        raise e
    finally:
        local_pyiceberg_catalog.drop_table(table_name)


@pytest.mark.integration()
def test_pyiceberg_written_catalog(local_iceberg_catalog):
    catalog_name, local_pyiceberg_catalog = local_iceberg_catalog
    with table_written_by_pyiceberg(local_pyiceberg_catalog) as catalog_table_name:
        df = daft.read_table(f"{catalog_name}.{catalog_table_name}")
        daft_pandas = df.to_pandas()
        iceberg_pandas = local_pyiceberg_catalog.load_table(catalog_table_name).scan().to_arrow().to_pandas()
        assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
@pytest.mark.skip
def test_daft_written_catalog(local_iceberg_catalog):
    catalog_name, local_pyiceberg_catalog = local_iceberg_catalog
    with table_written_by_daft(local_pyiceberg_catalog) as catalog_table_name:
        df = daft.read_table(f"{catalog_name}.{catalog_table_name}")
        daft_pandas = df.to_pandas()
        iceberg_pandas = local_pyiceberg_catalog.load_table(catalog_table_name).scan().to_arrow().to_pandas()
        assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])



from __future__ import annotations

import pytest


def test_simple_concat(make_df, with_morsel_size):
    df1 = make_df({"foo": [1, 2, 3]})
    df2 = make_df({"foo": [4, 5, 6]})
    result = df1.concat(df2)
    assert result.to_pydict() == {"foo": [1, 2, 3, 4, 5, 6]}


def test_concat_schema_mismatch(make_df, with_morsel_size):
    df1 = make_df({"foo": [1, 2, 3]})
    df2 = make_df({"foo": ["4", "5", "6"]})
    with pytest.raises(ValueError):
        df1.concat(df2)


def test_self_concat(make_df, with_morsel_size):
    df = make_df({"foo": [1, 2, 3]})
    assert df.concat(df).to_pydict() == {"foo": [1, 2, 3, 1, 2, 3]}



from __future__ import annotations

import daft.daft as native
from daft.expressions import Expression, list_


def monotonically_increasing_id() -> Expression:
    """Generates a column of monotonically increasing unique ids.

    The implementation puts the partition number in the upper 28 bits, and the row number in each partition
    in the lower 36 bits. This allows for 2^28 ≈ 268 million partitions and 2^40 ≈ 68 billion rows per partition.

    Returns:
        Expression: An expression that generates monotonically increasing IDs

    Examples:
        >>> import daft
        >>> from daft.functions import monotonically_increasing_id
        >>> daft.context.set_runner_ray()  # doctest: +SKIP
        >>>
        >>> df = daft.from_pydict({"a": [1, 2, 3, 4]}).into_partitions(2)
        >>> df = df.with_column("id", monotonically_increasing_id())
        >>> df.show()  # doctest: +SKIP
        ╭───────┬─────────────╮
        │ a     ┆ id          │
        │ ---   ┆ ---         │
        │ Int64 ┆ UInt64      │
        ╞═══════╪═════════════╡
        │ 1     ┆ 0           │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2     ┆ 1           │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 3     ┆ 68719476736 │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 4     ┆ 68719476737 │
        ╰───────┴─────────────╯
        <BLANKLINE>
        (Showing first 4 of 4 rows)

    """
    return Expression._from_pyexpr(native.monotonically_increasing_id())


def columns_sum(*exprs: Expression | str) -> Expression:
    """Sum values across columns.

    Args:
        exprs: The columns to sum.

    Examples:
        >>> import daft
        >>> from daft.functions import columns_sum
        >>> df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> df = df.with_column("sum", columns_sum("a", "b"))
        >>> df.show()
        ╭───────┬───────┬───────╮
        │ a     ┆ b     ┆ sum   │
        │ ---   ┆ ---   ┆ ---   │
        │ Int64 ┆ Int64 ┆ Int64 │
        ╞═══════╪═══════╪═══════╡
        │ 1     ┆ 4     ┆ 5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 2     ┆ 5     ┆ 7     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 3     ┆ 6     ┆ 9     │
        ╰───────┴───────┴───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    if not exprs:
        raise ValueError("columns_sum requires at least one expression")
    return list_(*exprs).list.sum().alias("columns_sum")


def columns_mean(*exprs: Expression | str) -> Expression:
    """Average values across columns. Akin to `columns_avg`.

    Args:
        exprs: The columns to average.

    Examples:
        >>> import daft
        >>> from daft.functions import columns_mean
        >>> df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> df = df.with_column("mean", columns_mean("a", "b"))
        >>> df.show()
        ╭───────┬───────┬─────────╮
        │ a     ┆ b     ┆ mean    │
        │ ---   ┆ ---   ┆ ---     │
        │ Int64 ┆ Int64 ┆ Float64 │
        ╞═══════╪═══════╪═════════╡
        │ 1     ┆ 4     ┆ 2.5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
        │ 2     ┆ 5     ┆ 3.5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
        │ 3     ┆ 6     ┆ 4.5     │
        ╰───────┴───────┴─────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    if not exprs:
        raise ValueError("columns_mean requires at least one expression")
    return list_(*exprs).list.mean().alias("columns_mean")


def columns_avg(*exprs: Expression | str) -> Expression:
    """Average values across columns. Akin to `columns_mean`.

    Args:
        exprs: The columns to average across.

    Examples:
        >>> import daft
        >>> from daft.functions import columns_avg
        >>> df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> df = df.with_column("avg", columns_avg("a", "b"))
        >>> df.show()
        ╭───────┬───────┬─────────╮
        │ a     ┆ b     ┆ avg     │
        │ ---   ┆ ---   ┆ ---     │
        │ Int64 ┆ Int64 ┆ Float64 │
        ╞═══════╪═══════╪═════════╡
        │ 1     ┆ 4     ┆ 2.5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
        │ 2     ┆ 5     ┆ 3.5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
        │ 3     ┆ 6     ┆ 4.5     │
        ╰───────┴───────┴─────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    if not exprs:
        raise ValueError("columns_avg requires at least one expression")
    return list_(*exprs).list.mean().alias("columns_avg")


def columns_min(*exprs: Expression | str) -> Expression:
    """Find the minimum value across columns.

    Args:
        exprs: The columns to find the minimum of.

    Examples:
        >>> import daft
        >>> from daft.functions import columns_min
        >>> df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> df = df.with_column("min", columns_min("a", "b"))
        >>> df.show()
        ╭───────┬───────┬───────╮
        │ a     ┆ b     ┆ min   │
        │ ---   ┆ ---   ┆ ---   │
        │ Int64 ┆ Int64 ┆ Int64 │
        ╞═══════╪═══════╪═══════╡
        │ 1     ┆ 4     ┆ 1     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 2     ┆ 5     ┆ 2     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 3     ┆ 6     ┆ 3     │
        ╰───────┴───────┴───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    if not exprs:
        raise ValueError("columns_min requires at least one expression")
    return list_(*exprs).list.min().alias("columns_min")


def columns_max(*exprs: Expression | str) -> Expression:
    """Find the maximum value across columns.

    Args:
        exprs: The columns to find the maximum of.

    Examples:
        >>> import daft
        >>> from daft.functions import columns_max
        >>> df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> df = df.with_column("max", columns_max("a", "b"))
        >>> df.show()
        ╭───────┬───────┬───────╮
        │ a     ┆ b     ┆ max   │
        │ ---   ┆ ---   ┆ ---   │
        │ Int64 ┆ Int64 ┆ Int64 │
        ╞═══════╪═══════╪═══════╡
        │ 1     ┆ 4     ┆ 4     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 2     ┆ 5     ┆ 5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 3     ┆ 6     ┆ 6     │
        ╰───────┴───────┴───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    if not exprs:
        raise ValueError("columns_max requires at least one expression")
    return list_(*exprs).list.max().alias("columns_max")


def row_number() -> Expression:
    """Return the row number of the current row (used for window functions).

    Examples:
        >>> import daft
        >>> from daft.window import Window
        >>> from daft.functions import row_number
        >>> df = daft.from_pydict({"category": ["A", "A", "A", "A", "B", "B", "B", "B"], "value": [1, 7, 2, 9, 1, 3, 3, 7]})
        >>>
        >>> # Ascending order
        >>> window = Window().partition_by("category").order_by("value")
        >>> df = df.with_column("row", row_number().over(window))
        >>> df = df.sort("category")
        >>> df.show()
        ╭──────────┬───────┬────────╮
        │ category ┆ value ┆ row    │
        │ ---      ┆ ---   ┆ ---    │
        │ Utf8     ┆ Int64 ┆ UInt64 │
        ╞══════════╪═══════╪════════╡
        │ A        ┆ 1     ┆ 1      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ A        ┆ 2     ┆ 2      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ A        ┆ 7     ┆ 3      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ A        ┆ 9     ┆ 4      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ B        ┆ 1     ┆ 1      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ B        ┆ 3     ┆ 2      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ B        ┆ 3     ┆ 3      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ B        ┆ 7     ┆ 4      │
        ╰──────────┴───────┴────────╯
        <BLANKLINE>
        (Showing first 8 rows)

    Returns:
        Expression: An expression that returns the row number of the current row.
    """
    return Expression._from_pyexpr(native.row_number())


def rank() -> Expression:
    """Return the rank of the current row (used for window functions).

    Examples:
        >>> import daft
        >>> from daft.window import Window
        >>> from daft.functions import rank
        >>> df = daft.from_pydict({"category": ["A", "A", "A", "A", "B", "B", "B", "B"], "value": [1, 3, 3, 7, 7, 7, 4, 4]})
        >>>
        >>> window = Window().partition_by("category").order_by("value", desc=True)
        >>> df = df.with_column("rank", rank().over(window))
        >>> df = df.sort("category")
        >>> df.show()
        ╭──────────┬───────┬────────╮
        │ category ┆ value ┆ rank   │
        │ ---      ┆ ---   ┆ ---    │
        │ Utf8     ┆ Int64 ┆ UInt64 │
        ╞══════════╪═══════╪════════╡
        │ A        ┆ 7     ┆ 1      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ A        ┆ 3     ┆ 2      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ A        ┆ 3     ┆ 2      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ A        ┆ 1     ┆ 4      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ B        ┆ 7     ┆ 1      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ B        ┆ 7     ┆ 1      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ B        ┆ 4     ┆ 3      │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ B        ┆ 4     ┆ 3      │
        ╰──────────┴───────┴────────╯
        <BLANKLINE>
        (Showing first 8 rows)

    Returns:
        Expression: An expression that returns the rank of the current row.
    """
    return Expression._from_pyexpr(native.rank())


def dense_rank() -> Expression:
    """Return the dense rank of the current row (used for window functions).

    The dense rank is the rank of the current row without gaps.

    Examples:
        >>> import daft
        >>> from daft.window import Window
        >>> from daft.functions import dense_rank
        >>> df = daft.from_pydict({"category": ["A", "A", "A", "A", "B", "B", "B", "B"], "value": [1, 3, 3, 7, 7, 7, 4, 4]})
        >>>
        >>> window = Window().partition_by("category").order_by("value", desc=True)
        >>> df = df.with_column("dense_rank", dense_rank().over(window))
        >>> df = df.sort("category")
        >>> df.show()
        ╭──────────┬───────┬────────────╮
        │ category ┆ value ┆ dense_rank │
        │ ---      ┆ ---   ┆ ---        │
        │ Utf8     ┆ Int64 ┆ UInt64     │
        ╞══════════╪═══════╪════════════╡
        │ A        ┆ 7     ┆ 1          │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ A        ┆ 3     ┆ 2          │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ A        ┆ 3     ┆ 2          │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ A        ┆ 1     ┆ 3          │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ B        ┆ 7     ┆ 1          │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ B        ┆ 7     ┆ 1          │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ B        ┆ 4     ┆ 2          │
        ├╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ B        ┆ 4     ┆ 2          │
        ╰──────────┴───────┴────────────╯
        <BLANKLINE>
        (Showing first 8 rows)

    Returns:
        Expression: An expression that returns the dense rank of the current row.
    """
    return Expression._from_pyexpr(native.dense_rank())



# fmt: off
# ^ ruff formatting breaks tests which are sensitive to format.

from __future__ import annotations

import daft
from daft.dataframe import DataFrame
from daft.dataframe.preview import PreviewFormat, PreviewFormatter


def test_show_default(make_df, valid_data):
    df: DataFrame = make_df(valid_data)
    df_preview = df._construct_show_preview(8)

    assert len(df_preview.partition) == len(valid_data)

    # Limit was greater than the actual total rows, so cap at the valid length
    assert df_preview.total_rows == 3


def test_show_some(make_df, valid_data, data_source):
    df = make_df(valid_data)
    df_preview = df._construct_show_preview(1)

    # Check the partition was trimmed to the preview length.
    assert len(df_preview.partition) == 1

    # Limit is less than DataFrame length, so we only know full DataFrame length if it was loaded from memory, e.g. arrow.
    variant = data_source
    if variant == "parquet":
        assert df_preview.total_rows is None
    elif variant == "arrow":
        assert df_preview.total_rows == len(valid_data)


def test_show_from_cached_repr(make_df, valid_data):
    df = make_df(valid_data)
    df = df.collect()
    df.__repr__()
    collected_preview = df._preview
    df_preview = df._construct_show_preview(8)

    # Check that cached preview from df.__repr__() was used.
    assert df_preview is collected_preview

    # Check lengths are valid
    assert len(df_preview.partition) == len(valid_data)
    assert df_preview.total_rows == 3


def test_show_from_cached_repr_prefix(make_df, valid_data):
    df = make_df(valid_data)
    df = df.collect(3)
    df.__repr__()
    df_preview = df._construct_show_preview(2)

    # Check that a prefix of the cached preview from df.__repr__() was used, so dataframe_total_rows should be set.
    assert len(df_preview.partition) == 2
    assert df_preview.total_rows == 3


def test_show_not_from_cached_repr(make_df, valid_data, data_source):
    df = make_df(valid_data)
    df = df.collect(2)
    df.__repr__()
    collected_preview = df._preview
    df_preview = df._construct_show_preview(8)

    variant = data_source
    if variant == "parquet":
        # Cached preview from df.__repr__() is NOT USED because data was not materialized from parquet.
        assert df_preview != collected_preview
    elif variant == "arrow":
        # Cached preview from df.__repr__() is USED because data was materialized from arrow.
        assert df_preview == collected_preview

    # Check lengths are valid
    assert len(df_preview.partition) == len(valid_data)
    assert df_preview.total_rows == 3


###
# tests for .show() with formatting options
###


def show(df: DataFrame, format: PreviewFormat, **options):
    """Helper since .show() will print to stdout and have user messages."""
    return PreviewFormatter(df._preview, df.schema(), format, **options)._to_text()


def test_show_with_options():
    df = daft.from_pydict(
        {
            "A": [1, 2, 3, 4],
            "B": [1.5, 2.5, 3.5, 4.5],
            "C": [True, True, False, False],
            "D": [None, None, None, None],
        }
    )

    columns = [
        {
            "info": "units",
            "align": "right",
        },
        {
            "info": "kg",
            "align": "right",
        },
        {},
        {},
    ]

    assert (
        show(df, format="markdown")
        == """
| A | B   | C     | D    |
|---|-----|-------|------|
| 1 | 1.5 | true  | None |
| 2 | 2.5 | true  | None |
| 3 | 3.5 | false | None |
| 4 | 4.5 | false | None |"""[1:]
    )

    assert (
        show(df, format="markdown", verbose=True)
        == """
| A (Int64) | B (Float64) | C (Boolean) | D (Null) |
|-----------|-------------|-------------|----------|
| 1         | 1.5         | true        | None     |
| 2         | 2.5         | true        | None     |
| 3         | 3.5         | false       | None     |
| 4         | 4.5         | false       | None     |"""[1:]
    )

    assert (
        show(df, format="markdown", verbose=True, columns=columns)
        == """
| A (units) | B (kg) | C (Boolean) | D (Null) |
|-----------|--------|-------------|----------|
|         1 |    1.5 | true        | None     |
|         2 |    2.5 | true        | None     |
|         3 |    3.5 | false       | None     |
|         4 |    4.5 | false       | None     |"""[1:]
    )


def test_show_with_wide_columns():
    df = daft.from_pydict(
        {
            "A": ["This is a very long text that exceeds the default max_width."],
            "B": ["Another extremely long piece of text that also exceeds the default max_width."],
        }
    )

    # default max_width
    assert show(df, format="markdown") == """
| A                            | B                            |
|------------------------------|------------------------------|
| This is a very long text th… | Another extremely long piec… |"""[1:]

    # explicit max_width
    assert show(df, format="markdown", max_width=12) == """
| A          | B          |
|------------|------------|
| This is a… | Another e… |"""[1:]

    # no max_width
    assert show(df, format="simple", max_width=None) == """
 A                                                              B                                                                             
--------------------------------------------------------------+-------------------------------------------------------------------------------
 This is a very long text that exceeds the default max_width.   Another extremely long piece of text that also exceeds the default max_width. """[1:]


def test_show_with_many_columns():
    df = daft.from_pylist(
        [
            {
                "A": 1,
                "B": 2,
                "C": 3,
                "D": 4,
                "E": 5,
                "F": 6,
                "G": 7,
                "H": 8,
                "I": 9,
                "J": 10,
                "K": 11,
                "L": 12,
                "M": 13,
                "N": 14,
                "O": 15,
                "P": 16,
                "Q": 17,
                "R": 18,
                "S": 19,
                "T": 20,
                "U": 21,
                "V": 22,
                "W": 23,
                "X": 24,
                "Y": 25,
                "Z": 26,
            }
        ]
    )
    assert show(df, format="simple") == """
 A   B   C   D   E   F   G   H   I   J    K    L    M    N    O    P    Q    R    S    T    U    V    W    X    Y    Z  
---+---+---+---+---+---+---+---+---+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----
 1   2   3   4   5   6   7   8   9   10   11   12   13   14   15   16   17   18   19   20   21   22   23   24   25   26 """[1:]



from __future__ import annotations

from itertools import product

import pytest

import daft
from daft.expressions.expressions import _resolved_col
from daft.recordbatch import MicroPartition

PRED_PUSHDOWN_FILES = [
    "s3://daft-public-data/test_fixtures/parquet-dev/sampled-tpch-with-stats.parquet",
    "tests/assets/parquet-data/sampled-tpch-with-stats.parquet",
]


@pytest.mark.integration()
@pytest.mark.parametrize(
    "path, pred, limit",
    product(
        PRED_PUSHDOWN_FILES,
        [_resolved_col("L_ORDERKEY") == 7, _resolved_col("L_ORDERKEY") == 10000, daft.lit(True)],
        [None, 1, 1000],
    ),
)
def test_parquet_filter_pushdowns(path, pred, limit, aws_public_s3_config):
    with_pushdown = MicroPartition.read_parquet(path, predicate=pred, num_rows=limit, io_config=aws_public_s3_config)
    after = MicroPartition.read_parquet(path, io_config=aws_public_s3_config).filter([pred])
    if limit is not None:
        after = after.head(limit)
    assert with_pushdown.to_arrow() == after.to_arrow()


@pytest.mark.integration()
@pytest.mark.parametrize(
    "path, pred",
    product(
        PRED_PUSHDOWN_FILES, [_resolved_col("L_ORDERKEY") == 7, _resolved_col("L_ORDERKEY") == 10000, daft.lit(True)]
    ),
)
def test_parquet_filter_pushdowns_disjoint_predicate(path, pred, aws_public_s3_config):
    with_pushdown = MicroPartition.read_parquet(
        path, predicate=pred, columns=["L_QUANTITY"], io_config=aws_public_s3_config
    )
    after = (
        MicroPartition.read_parquet(path, io_config=aws_public_s3_config)
        .filter([pred])
        .eval_expression_list([_resolved_col("L_QUANTITY")])
    )
    assert with_pushdown.to_arrow() == after.to_arrow()


@pytest.mark.integration()
@pytest.mark.parametrize(
    "path, pred",
    product(
        ["tests/assets/parquet-data/mvp.parquet", "s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet"],
        [_resolved_col("a") == 7, _resolved_col("a") == 10000, daft.lit(True)],
    ),
)
def test_parquet_filter_pushdowns_disjoint_predicate_no_stats(path, pred, aws_public_s3_config):
    with_pushdown = MicroPartition.read_parquet(path, predicate=pred, columns=["b"], io_config=aws_public_s3_config)
    after = (
        MicroPartition.read_parquet(path, io_config=aws_public_s3_config)
        .filter([pred])
        .eval_expression_list([_resolved_col("b")])
    )
    assert with_pushdown.to_arrow() == after.to_arrow()



from __future__ import annotations

import pytest

from daft.expressions import col, lit
from daft.recordbatch import MicroPartition


@pytest.mark.parametrize(
    "input_a,input_b,expected_result",
    [
        # Basic ASCII concatenation
        (
            [b"Hello", b"Test", b"", b"End"],
            [b" World", b"ing", b"Empty", b"!"],
            [b"Hello World", b"Testing", b"Empty", b"End!"],
        ),
        # Special binary sequences
        (
            [
                b"\x00\x01",  # Null and control chars
                b"\xff\xfe",  # High-value bytes
                b"Hello\x00",  # String with null
                b"\xe2\x98",  # Partial UTF-8
                b"\xf0\x9f\x98",  # Another partial UTF-8
            ],
            [
                b"\x02\x03",  # More control chars
                b"\xfd\xfc",  # More high-value bytes
                b"\x00World",  # Null and string
                b"\x83",  # Complete the UTF-8 snowman
                b"\x89",  # Complete the UTF-8 winking face
            ],
            [
                b"\x00\x01\x02\x03",  # Concatenated control chars
                b"\xff\xfe\xfd\xfc",  # Concatenated high-value bytes
                b"Hello\x00\x00World",  # String with multiple nulls
                b"\xe2\x98\x83",  # Complete UTF-8 snowman (☃)
                b"\xf0\x9f\x98\x89",  # Complete UTF-8 winking face (😉)
            ],
        ),
        # Nulls and empty strings
        (
            [b"Hello", None, b"", b"Test", None, b"End", b""],
            [b" World", b"!", None, None, b"ing", b"", b"Empty"],
            [b"Hello World", None, None, None, None, b"End", b"Empty"],
        ),
        # Mixed length concatenation
        (
            [b"a", b"ab", b"abc", b"abcd"],
            [b"1", b"12", b"123", b"1234"],
            [b"a1", b"ab12", b"abc123", b"abcd1234"],
        ),
        # Empty string combinations
        (
            [b"", b"", b"Hello", b"World", b""],
            [b"", b"Test", b"", b"", b"!"],
            [b"", b"Test", b"Hello", b"World", b"!"],
        ),
        # Complex UTF-8 sequences
        (
            [
                b"\xe2\x98\x83",  # Snowman
                b"\xf0\x9f\x98\x89",  # Winking face
                b"\xf0\x9f\x8c\x88",  # Rainbow
                b"\xe2\x98\x83\xf0\x9f\x98\x89",  # Snowman + Winking face
            ],
            [
                b"\xf0\x9f\x98\x89",  # Winking face
                b"\xe2\x98\x83",  # Snowman
                b"\xe2\x98\x83",  # Snowman
                b"\xf0\x9f\x8c\x88",  # Rainbow
            ],
            [
                b"\xe2\x98\x83\xf0\x9f\x98\x89",  # Snowman + Winking face
                b"\xf0\x9f\x98\x89\xe2\x98\x83",  # Winking face + Snowman
                b"\xf0\x9f\x8c\x88\xe2\x98\x83",  # Rainbow + Snowman
                b"\xe2\x98\x83\xf0\x9f\x98\x89\xf0\x9f\x8c\x88",  # Snowman + Winking face + Rainbow
            ],
        ),
        # Zero bytes in different positions
        (
            [
                b"\x00abc",  # Leading zero
                b"abc\x00",  # Trailing zero
                b"ab\x00c",  # Middle zero
                b"\x00ab\x00c\x00",  # Multiple zeros
            ],
            [
                b"def\x00",  # Trailing zero
                b"\x00def",  # Leading zero
                b"d\x00ef",  # Middle zero
                b"\x00de\x00f\x00",  # Multiple zeros
            ],
            [
                b"\x00abcdef\x00",  # Zeros at ends
                b"abc\x00\x00def",  # Adjacent zeros
                b"ab\x00cd\x00ef",  # Separated zeros
                b"\x00ab\x00c\x00\x00de\x00f\x00",  # Many zeros
            ],
        ),
    ],
)
def test_binary_concat(
    input_a: list[bytes | None], input_b: list[bytes | None], expected_result: list[bytes | None]
) -> None:
    table = MicroPartition.from_pydict({"a": input_a, "b": input_b})
    result = table.eval_expression_list([col("a").binary.concat(col("b"))])
    assert result.to_pydict() == {"a": expected_result}


@pytest.mark.parametrize(
    "input_data,literal,expected_result",
    [
        # Basic broadcasting
        (
            [b"Hello", b"Goodbye", b"Test"],
            b" World!",
            [b"Hello World!", b"Goodbye World!", b"Test World!"],
        ),
        # Broadcasting with nulls
        (
            [b"Hello", None, b"Test"],
            b" World!",
            [b"Hello World!", None, b"Test World!"],
        ),
        # Broadcasting with special sequences
        (
            [b"\x00\x01", b"\xff\xfe", b"Hello\x00"],
            b"\x02\x03",
            [b"\x00\x01\x02\x03", b"\xff\xfe\x02\x03", b"Hello\x00\x02\x03"],
        ),
        # Broadcasting with empty strings
        (
            [b"", b"Test", b""],
            b"\xff\xfe",
            [b"\xff\xfe", b"Test\xff\xfe", b"\xff\xfe"],
        ),
        # Broadcasting with UTF-8
        (
            [b"Hello", b"Test", b"Goodbye"],
            b"\xe2\x98\x83",  # Snowman
            [b"Hello\xe2\x98\x83", b"Test\xe2\x98\x83", b"Goodbye\xe2\x98\x83"],
        ),
        # Broadcasting with zero bytes
        (
            [b"Hello", b"Test\x00", b"\x00World"],
            b"\x00",
            [b"Hello\x00", b"Test\x00\x00", b"\x00World\x00"],
        ),
        # Broadcasting with literal None
        (
            [b"Hello", None, b"Test", b""],
            None,
            [None, None, None, None],  # Any concat with None should result in None
        ),
    ],
)
def test_binary_concat_broadcast(
    input_data: list[bytes | None], literal: bytes | None, expected_result: list[bytes | None]
) -> None:
    # Test right-side broadcasting
    table = MicroPartition.from_pydict({"a": input_data})
    result = table.eval_expression_list([col("a").binary.concat(literal)])
    assert result.to_pydict() == {"a": expected_result}

    # Test left-side broadcasting
    table = MicroPartition.from_pydict({"b": input_data})
    result = table.eval_expression_list([lit(literal).binary.concat(col("b"))])
    if literal is None:
        # When literal is None, all results should be None
        assert result.to_pydict() == {"literal": [None] * len(input_data)}
    else:
        assert result.to_pydict() == {
            "literal": [
                lit + data if data is not None else None for lit, data in zip([literal] * len(input_data), input_data)
            ]
        }


def test_binary_concat_edge_cases() -> None:
    # Test various edge cases
    table = MicroPartition.from_pydict(
        {
            "a": [
                b"",  # Empty string
                b"\x00",  # Single null byte
                b"\xff",  # Single high byte
                b"Hello",  # Normal string
                None,  # Null value
                b"\xe2\x98\x83",  # UTF-8 sequence
                b"\xf0\x9f\x98\x89",  # Another UTF-8 sequence
                b"\x80\x81\x82",  # Binary sequence
                b"\xff\xff\xff",  # High bytes
            ],
            "b": [
                b"",  # Empty + Empty
                b"\x00",  # Null + Null
                b"\x00",  # High + Null
                b"",  # Normal + Empty
                None,  # Null + Null
                b"\xf0\x9f\x98\x89",  # UTF-8 + UTF-8
                b"\xe2\x98\x83",  # UTF-8 + UTF-8
                b"\x83\x84\x85",  # Binary + Binary
                b"\xfe\xfe\xfe",  # High bytes + High bytes
            ],
        }
    )
    result = table.eval_expression_list([col("a").binary.concat(col("b"))])
    assert result.to_pydict() == {
        "a": [
            b"",  # Empty + Empty = Empty
            b"\x00\x00",  # Null + Null = Two nulls
            b"\xff\x00",  # High + Null = High then null
            b"Hello",  # Normal + Empty = Normal
            None,  # Null + Null = Null
            b"\xe2\x98\x83\xf0\x9f\x98\x89",  # Snowman + Winking face
            b"\xf0\x9f\x98\x89\xe2\x98\x83",  # Winking face + Snowman
            b"\x80\x81\x82\x83\x84\x85",  # Binary sequence concatenation
            b"\xff\xff\xff\xfe\xfe\xfe",  # High bytes concatenation
        ]
    }


def test_binary_concat_errors() -> None:
    # Test concat with incompatible type (string)
    table = MicroPartition.from_pydict({"a": [b"hello", b"world"], "b": ["foo", "bar"]})
    with pytest.raises(Exception, match="Expects inputs to concat to be binary, but received a#Binary and b#Utf8"):
        table.eval_expression_list([col("a").binary.concat(col("b"))])

    # Test concat with incompatible type (integer)
    table = MicroPartition.from_pydict({"a": [b"hello", b"world"], "b": [1, 2]})
    with pytest.raises(Exception, match="Expects inputs to concat to be binary, but received a#Binary and b#Int64"):
        table.eval_expression_list([col("a").binary.concat(col("b"))])

    # Test concat with incompatible type (float)
    table = MicroPartition.from_pydict({"a": [b"hello", b"world"], "b": [1.0, 2.0]})
    with pytest.raises(Exception, match="Expects inputs to concat to be binary, but received a#Binary and b#Float64"):
        table.eval_expression_list([col("a").binary.concat(col("b"))])

    # Test concat with incompatible type (boolean)
    table = MicroPartition.from_pydict({"a": [b"hello", b"world"], "b": [True, False]})
    with pytest.raises(Exception, match="Expects inputs to concat to be binary, but received a#Binary and b#Boolean"):
        table.eval_expression_list([col("a").binary.concat(col("b"))])

    # Test concat with wrong number of arguments
    table = MicroPartition.from_pydict({"a": [b"hello", b"world"], "b": [b"foo", b"bar"], "c": [b"test", b"data"]})
    with pytest.raises(
        Exception, match="(?:ExpressionBinaryNamespace.)?concat\\(\\) takes 2 positional arguments but 3 were given"
    ):
        table.eval_expression_list([col("a").binary.concat(col("b"), col("c"))])

    # Test concat with no arguments
    table = MicroPartition.from_pydict({"a": [b"hello", b"world"]})
    with pytest.raises(
        Exception, match="(?:ExpressionBinaryNamespace.)?concat\\(\\) missing 1 required positional argument: 'other'"
    ):
        table.eval_expression_list([col("a").binary.concat()])



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import Request, urlopen

from deltalake.table import DeltaTable

import daft
import daft.exceptions
from daft.daft import (
    FileFormatConfig,
    ParquetSourceConfig,
    PyPartitionField,
    PyPushdowns,
    S3Config,
    ScanTask,
    StorageConfig,
)
from daft.io.object_store_options import io_config_to_storage_options
from daft.io.scan import ScanOperator
from daft.logical.schema import Schema

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

logger = logging.getLogger(__name__)


def get_s3_bucket_region(bucket_name: str) -> str | None:
    # When making a https request to https://{bucket_name}.s3.amazonaws.com/, aws returns either a 200 response, 403 response, or 404 response.
    # In the header of the 200 and 403 responses, there is an `x-amz-bucket-region` field from which we can extract the bucket's region.
    url = f"https://{bucket_name}.s3.amazonaws.com"

    try:
        req = Request(url, method="HEAD")
        with urlopen(req) as response:
            return response.headers.get("x-amz-bucket-region")
    except HTTPError as e:
        bucket_region = e.headers.get("x-amz-bucket-region")
        if bucket_region is None:
            logger.warning(
                "Failed to get the S3 bucket region using the given S3 uri. HTTPError error: %s",
                e,
            )
        return bucket_region
    except Exception as e:
        logger.warning(
            "Failed to get the S3 bucket region using the given S3 uri. Error: %s",
            e,
        )
        return None


class DeltaLakeScanOperator(ScanOperator):
    def __init__(
        self, table_uri: str, storage_config: StorageConfig, version: int | str | datetime | None = None
    ) -> None:
        super().__init__()

        # Unfortunately delta-rs doesn't do very good inference of credentials for S3. Thus the current Daft behavior of passing
        # in `None` for credentials will cause issues when instantiating the DeltaTable without credentials.
        #
        # Thus, if we don't detect any credentials being available, we attempt to detect it from the environment using our Daft credentials chain.
        #
        # See: https://github.com/delta-io/delta-rs/issues/2117
        deltalake_sdk_io_config = storage_config.io_config
        scheme = urlparse(table_uri).scheme
        if scheme == "s3" or scheme == "s3a":
            # Try to get the bucket's region.
            if deltalake_sdk_io_config.s3.region_name is None:
                bucket_name = urlparse(table_uri).netloc
                region = get_s3_bucket_region(bucket_name)
                if region is not None:
                    deltalake_sdk_io_config = deltalake_sdk_io_config.replace(
                        s3=deltalake_sdk_io_config.s3.replace(region_name=region)
                    )

            # Try to get config from the environment
            if any([deltalake_sdk_io_config.s3.key_id is None, deltalake_sdk_io_config.s3.region_name is None]):
                try:
                    s3_config_from_env = S3Config.from_env()
                # Sometimes S3Config.from_env throws an error, for example on CI machines with weird metadata servers.
                except daft.exceptions.DaftCoreException:
                    pass
                else:
                    if (
                        deltalake_sdk_io_config.s3.key_id is None
                        and deltalake_sdk_io_config.s3.access_key is None
                        and deltalake_sdk_io_config.s3.session_token is None
                    ):
                        deltalake_sdk_io_config = deltalake_sdk_io_config.replace(
                            s3=deltalake_sdk_io_config.s3.replace(
                                key_id=s3_config_from_env.key_id,
                                access_key=s3_config_from_env.access_key,
                                session_token=s3_config_from_env.session_token,
                            )
                        )
                    if deltalake_sdk_io_config.s3.region_name is None:
                        deltalake_sdk_io_config = deltalake_sdk_io_config.replace(
                            s3=deltalake_sdk_io_config.s3.replace(
                                region_name=s3_config_from_env.region_name,
                            )
                        )
        elif scheme == "gcs" or scheme == "gs":
            # TO-DO: Handle any key-value replacements in `io_config` if there are missing elements
            pass
        elif scheme == "az" or scheme == "abfs" or scheme == "abfss":
            # TO-DO: Handle any key-value replacements in `io_config` if there are missing elements
            pass

        self._table = DeltaTable(
            table_uri, storage_options=io_config_to_storage_options(deltalake_sdk_io_config, table_uri)
        )

        if version is not None:
            self._table.load_as_version(version)

        self._storage_config = storage_config
        self._schema = Schema.from_pyarrow_schema(self._table.schema().to_pyarrow())
        partition_columns = set(self._table.metadata().partition_columns)
        self._partition_keys = [
            PyPartitionField(field._field) for field in self._schema if field.name in partition_columns
        ]

    def schema(self) -> Schema:
        return self._schema

    def name(self) -> str:
        return "DeltaLakeScanOperator"

    def display_name(self) -> str:
        return f"DeltaLakeScanOperator({self._table.metadata().name})"

    def partitioning_keys(self) -> list[PyPartitionField]:
        return self._partition_keys

    def multiline_display(self) -> list[str]:
        return [
            self.display_name(),
            f"Schema = {self._schema}",
            f"Partitioning keys = {self.partitioning_keys()}",
            # TODO(Clark): Improve repr of storage config here.
            f"Storage config = {self._storage_config}",
        ]

    def to_scan_tasks(self, pushdowns: PyPushdowns) -> Iterator[ScanTask]:
        import pyarrow as pa

        # TODO(Clark): Push limit and filter expressions into deltalake action fetch, to prune the files returned.
        # Issue: https://github.com/Eventual-Inc/Daft/issues/1953
        add_actions: pa.RecordBatch = self._table.get_add_actions()

        if len(self.partitioning_keys()) > 0 and pushdowns.partition_filters is None:
            logger.warning(
                "%s has partitioning keys = %s, but no partition filter was specified. This will result in a full table scan.",
                self.display_name(),
                self.partitioning_keys(),
            )

        # TODO(Clark): Add support for deletion vectors.
        # Issue: https://github.com/Eventual-Inc/Daft/issues/1954
        if "deletionVector" in add_actions.schema.names:
            raise NotImplementedError(
                "Delta Lake deletion vectors are not yet supported; please let the Daft team know if you'd like to see this feature!\n"
                "Deletion records can be dropped from this table to allow it to be read with Daft: https://docs.delta.io/latest/delta-drop-feature.html"
            )

        # TODO(Clark): Add support for column mappings.
        # Issue: https://github.com/Eventual-Inc/Daft/issues/1955

        limit_files = pushdowns.limit is not None and pushdowns.filters is None and pushdowns.partition_filters is None
        rows_left = pushdowns.limit if pushdowns.limit is not None else 0
        scan_tasks = []
        is_partitioned = (
            "partition_values" in add_actions.schema.names
            and add_actions.schema.field("partition_values").type.num_fields > 0
        )
        for task_idx in range(add_actions.num_rows):
            if limit_files and rows_left <= 0:
                break

            # NOTE: The paths in the transaction log consist of the post-table-uri suffix.
            path = os.path.join(self._table.table_uri, add_actions["path"][task_idx].as_py())

            try:
                record_count = add_actions["num_records"][task_idx].as_py()
            except KeyError:
                record_count = None

            try:
                size_bytes = add_actions["size_bytes"][task_idx].as_py()
            except KeyError:
                size_bytes = None
            file_format_config = FileFormatConfig.from_parquet_config(ParquetSourceConfig())

            if is_partitioned:
                dtype = add_actions.schema.field("partition_values").type
                part_values = add_actions["partition_values"][task_idx]
                arrays = {}
                for field_idx in range(dtype.num_fields):
                    field_name = dtype.field(field_idx).name
                    try:
                        arrow_arr = pa.array([part_values[field_name]], type=dtype.field(field_idx).type)
                    except (pa.ArrowInvalid, pa.ArrowTypeError, pa.ArrowNotImplementedError):
                        # pyarrow < 13.0.0 doesn't accept pyarrow scalars in the array constructor.
                        arrow_arr = pa.array([part_values[field_name].as_py()], type=dtype.field(field_idx).type)
                    arrays[field_name] = daft.Series.from_arrow(arrow_arr, field_name)
                partition_values = daft.recordbatch.RecordBatch.from_pydict(arrays)._recordbatch
            else:
                partition_values = None

            # Populate scan task with column-wise stats.
            schema_names = add_actions.schema.names
            if "min" in schema_names and "max" in schema_names:
                dtype = add_actions.schema.field("min").type
                min_values = add_actions["min"][task_idx]
                max_values = add_actions["max"][task_idx]
                # TODO(Clark): Add support for tracking null counts in column stats.
                # null_counts = add_actions["null_count"][task_idx]
                arrays = {}
                for field_idx in range(dtype.num_fields):
                    field_name = dtype.field(field_idx).name
                    try:
                        arrow_arr = pa.array(
                            [min_values[field_name], max_values[field_name]], type=dtype.field(field_idx).type
                        )
                    except (pa.ArrowInvalid, pa.ArrowTypeError, pa.ArrowNotImplementedError):
                        # pyarrow < 13.0.0 doesn't accept pyarrow scalars in the array constructor.
                        arrow_arr = pa.array(
                            [min_values[field_name].as_py(), max_values[field_name].as_py()],
                            type=dtype.field(field_idx).type,
                        )
                    arrays[field_name] = daft.Series.from_arrow(arrow_arr, field_name)
                stats = daft.recordbatch.RecordBatch.from_pydict(arrays)
            else:
                stats = None
            st = ScanTask.catalog_scan_task(
                file=path,
                file_format=file_format_config,
                schema=self._schema._schema,
                num_rows=record_count,
                storage_config=self._storage_config,
                size_bytes=size_bytes,
                iceberg_delete_files=None,
                pushdowns=pushdowns,
                partition_values=partition_values,
                stats=stats._recordbatch if stats is not None else None,
            )
            if st is None:
                continue
            if record_count is not None:
                rows_left -= record_count
            scan_tasks.append(st)
        return iter(scan_tasks)

    def can_absorb_filter(self) -> bool:
        return False

    def can_absorb_limit(self) -> bool:
        return False

    def can_absorb_select(self) -> bool:
        return True
